import re
from typing import Callable, Dict, List, Optional, Set, Type

from mypy.nodes import (
    AssignmentStmt,
    CallExpr,
    ClassDef,
    Decorator,
    MemberExpr,
    NameExpr,
    OverloadedFuncDef,
    SymbolTableNode,
    TypeAlias as MypyTypeAlias,
    TypeInfo,
    Var,
)
from mypy.options import Options
from mypy.plugin import AttributeContext, ClassDefContext, FunctionSigContext, MethodContext, Plugin
from mypy.types import AnyType, CallableType, FunctionLike, Instance, Type as MypyType, TypeAliasType, UnionType

TypeResolver = Callable[[str, List[MypyType]], MypyType]
NODE_ATTRIBUTE_REGEX = r"^[a-z].*$"
BASE_NODE_FULLNAME = "vellum.workflows.nodes.bases.base.BaseNode"

DESCRIPTOR_PATHS: list[tuple[str, str, str]] = [
    (
        "vellum.workflows.outputs.base.BaseOutputs",
        "vellum.workflows.references.output.OutputReference",
        r"^[^_].*$",
    ),
    (
        f"{BASE_NODE_FULLNAME}.ExternalInputs",
        "vellum.workflows.references.external_input.ExternalInputReference",
        r"^[^_].*$",
    ),
    (
        f"{BASE_NODE_FULLNAME}.Execution",
        "vellum.workflows.references.execution_count.ExecutionCountReference",
        r"^count$",
    ),
    (
        "vellum.workflows.state.base.BaseState",
        "vellum.workflows.references.state_value.StateValueReference",
        r"^[^_].*$",
    ),
    (
        "vellum.workflows.inputs.base.BaseInputs",
        "vellum.workflows.references.workflow_input.WorkflowInputReference",
        r"^[^_].*$",
    ),
    (
        BASE_NODE_FULLNAME,
        "vellum.workflows.references.node.NodeReference",
        NODE_ATTRIBUTE_REGEX,
    ),
]


def _is_subclass(type_info: Optional[TypeInfo], fullname: str) -> bool:
    if not type_info:
        return False

    if type_info.fullname == fullname:
        return True

    return any(base.type.fullname == fullname or _is_subclass(base.type, fullname) for base in type_info.bases)


def _get_attribute_mypy_type(type_info: TypeInfo, attribute_name: str) -> Optional[MypyType]:
    type_node = type_info.names.get(attribute_name)
    if type_node:
        return type_node.type

    bases = type_info.bases
    for base in bases:
        mypy_type = _get_attribute_mypy_type(base.type, attribute_name)
        if mypy_type:
            return mypy_type

    return None


class VellumMypyPlugin(Plugin):
    """
    This plugin is responsible for properly supporting types for all of the magic we
    do with Descriptors and this library in general.
    """

    def __init__(self, options: Options) -> None:
        """
        Mypy performs its analyses in two phases: semantic analysis and type checking.
        So we initialize state that we need to use across both phases here.

        - `_calls_with_nested_descriptor_expressions`: A set of instances that are defined
          within a BaseNode that reference a descriptor. Classes (e.g. Dataclasses) don't
          support Descriptors by default, so we first need to ensure we're in the context of
          a `BaseNode`, before then editing the signature to support Descriptors.

        - `_nested_descriptor_expressions`: A mapping of MemberExprs that point to a
          Descriptor to a callable that will resolve the type of the Descriptor. This field
          is used in combination with `_calls_with_nested_descriptor_expressions` to
          edit the signature of the function when a Descriptor is used as an argument.
        """

        self._calls_with_nested_descriptor_expressions: Set[CallExpr] = set()
        self._nested_descriptor_expressions: Dict[MemberExpr, TypeResolver] = {}

        super().__init__(options)

    def get_class_attribute_hook(self, fullname: str) -> Optional[Callable[[AttributeContext], MypyType]]:
        """
        This hook is used whenever we're accessing an attribute of a class. e.g. `MyClass.my_attribute`.

        We use it to replace all special class attribute references with our descriptors.

        TODO: We still need to support all other descriptors besides Outputs.
        https://app.shortcut.com/vellum/story/4768
        """
        return self._class_attribute_hook

    def _class_attribute_hook(self, ctx: AttributeContext) -> MypyType:
        if not isinstance(ctx.type, CallableType):
            return ctx.default_attr_type

        if not isinstance(ctx.type.ret_type, Instance):
            return ctx.default_attr_type

        if not isinstance(ctx.context, MemberExpr):
            return ctx.default_attr_type

        for base_path, descriptor_path, attribute_regex in DESCRIPTOR_PATHS:
            if not re.match(attribute_regex, ctx.context.name):
                continue

            if not _is_subclass(ctx.type.ret_type.type, base_path):
                continue

            symbol = ctx.type.ret_type.type.names.get(ctx.context.name)
            if symbol and isinstance(symbol.node, (Decorator, OverloadedFuncDef)):
                continue

            info = self.lookup_fully_qualified(descriptor_path)
            if info and isinstance(info.node, TypeInfo):
                return Instance(info.node, [self._resolve_descriptor_type(ctx.default_attr_type)])

        return ctx.default_attr_type

    def get_base_class_hook(self, fullname: str) -> Optional[Callable[[ClassDefContext], None]]:
        """
        This hook is used whenever we're defining a class. e.g. `class MyClass(BaseNode): ...`.

        We add support for any special Base Class we define in our project.
        """

        return self._base_class_hook

    def _base_class_hook(self, ctx: ClassDefContext) -> None:
        if _is_subclass(ctx.cls.info, "vellum.workflows.nodes.core.templating_node.node.TemplatingNode"):
            self._dynamic_output_node_class_hook(ctx, "result")
        elif _is_subclass(
            ctx.cls.info, "vellum.workflows.nodes.displayable.code_execution_node.node.CodeExecutionNode"
        ):
            self._dynamic_output_node_class_hook(ctx, "result")
        elif _is_subclass(ctx.cls.info, "vellum.workflows.nodes.displayable.final_output_node.node.FinalOutputNode"):
            self._dynamic_output_node_class_hook(ctx, "value")

        if _is_subclass(ctx.cls.info, BASE_NODE_FULLNAME):
            return self._base_node_class_hook(ctx)

        if _is_subclass(ctx.cls.info, "vellum.workflows.workflows.base.BaseWorkflow"):
            return self._base_workflow_class_hook(ctx)

        if _is_subclass(ctx.cls.info, "vellum.workflows.outputs.base.BaseOutputs"):
            return self._base_node_outputs_class_hook(ctx)

    def _dynamic_output_node_class_hook(self, ctx: ClassDefContext, attribute_name: str) -> None:
        """
        We use this hook to properly annotate the Outputs class for Templating Node and
        Code Execution Node using the resolved type
        of the TemplatingNode's and CodeExecutionNode's class _OutputType generic.
        """
        node_info = ctx.cls.info
        node_bases = ctx.cls.info.bases
        if not node_bases:
            return
        if not isinstance(node_bases[0], Instance):
            return

        base_args = node_bases[0].args
        base_node = node_bases[0].type

        if len(base_args) != 2:
            return

        base_node_resolved_type = base_args[1]
        if isinstance(base_node_resolved_type, AnyType):
            base_node_resolved_type = ctx.api.named_type("builtins.str")

        base_node_outputs = base_node.names.get("Outputs")
        if not base_node_outputs:
            return

        current_node_outputs = node_info.names.get("Outputs")
        if not current_node_outputs and isinstance(base_node_outputs.node, TypeInfo):
            node_info.names["Outputs"] = base_node_outputs.copy()

            new_outputs_fullname = f"{node_info.fullname}.Outputs"
            new_outputs_defn_raw = base_node_outputs.node.defn.serialize()
            new_outputs_defn_raw["fullname"] = new_outputs_fullname
            new_outputs_defn = ClassDef.deserialize(new_outputs_defn_raw)
            new_outputs_sym = TypeInfo(
                names=base_node_outputs.node.names.copy(),
                defn=new_outputs_defn,
                module_name=node_info.module_name,
            )

            base_result_sym = base_node_outputs.node.names[attribute_name].node
            if isinstance(base_result_sym, Var):
                new_result_sym = Var.deserialize(base_result_sym.serialize())
                new_result_sym._fullname = f"{new_outputs_fullname}.{attribute_name}"
                new_result_sym.type = base_node_resolved_type
                new_outputs_sym.names[attribute_name].node = new_result_sym
                new_outputs_sym.bases.append(Instance(base_node_outputs.node, []))
                node_info.names["Outputs"].node = new_outputs_sym

    def _base_node_class_hook(self, ctx: ClassDefContext) -> None:
        """
        Special handling of BaseNode class definitions
        """
        self._redefine_class_attributes_with_descriptors(ctx)

    def _base_node_outputs_class_hook(self, ctx: ClassDefContext) -> None:
        """
        Special handling of BaseNode.Outputs class definitions
        """
        self._redefine_class_attributes_with_descriptors(ctx)

    def _redefine_class_attributes_with_descriptors(self, ctx: ClassDefContext) -> None:
        """
        Given a class definition, we want to redefine all of the class attributes to accept both
        the original defined type, and the descriptor version of the type.
        """

        for key, sym in ctx.cls.info.names.items():
            if ctx.cls.info.fullname == BASE_NODE_FULLNAME:
                continue

            if not isinstance(sym.node, Var):
                continue

            if not re.match(NODE_ATTRIBUTE_REGEX, key):
                continue

            type_ = sym.node.type
            if not type_:
                continue

            sym.node.type = self._get_resolvable_type(
                lambda fullname, types: ctx.api.named_type(fullname, types), type_
            )

        # Supports descriptors assigned to nested classes
        type_resolver = lambda fullname, types: ctx.api.named_type(fullname, types)  # noqa: E731
        for assignment in ctx.cls.defs.body:
            if not isinstance(assignment, AssignmentStmt):
                continue

            call_expr = assignment.rvalue
            if not isinstance(call_expr, CallExpr):
                continue

            self._collect_descriptor_expressions(type_resolver, call_expr)

    def _collect_descriptor_expressions(self, type_resolver: TypeResolver, call_expr: CallExpr) -> None:
        for arg in call_expr.args:
            if isinstance(arg, CallExpr):
                self._collect_descriptor_expressions(type_resolver, arg)
                continue

            if not isinstance(arg, MemberExpr):
                continue

            is_arg_registered = False
            for base_path, _, attribute_regex in DESCRIPTOR_PATHS:
                if not re.match(attribute_regex, arg.name):
                    continue

                if not isinstance(arg.expr, (NameExpr, MemberExpr)):
                    continue

                if not isinstance(arg.expr.node, TypeInfo):
                    continue

                if not _is_subclass(arg.expr.node, base_path):
                    continue

                is_arg_registered = True
                self._calls_with_nested_descriptor_expressions.add(call_expr)
                self._nested_descriptor_expressions[arg] = type_resolver
                break

            if not is_arg_registered:
                # Need to check node outputs that are inherited
                if (
                    re.match(r"^[^_].*$", arg.name)
                    and isinstance(arg.expr, MemberExpr)
                    and isinstance(arg.expr.expr, NameExpr)
                    and isinstance(arg.expr.expr.node, TypeInfo)
                    and _is_subclass(arg.expr.expr.node, BASE_NODE_FULLNAME)
                    and arg.expr.name == "Outputs"
                ):
                    self._calls_with_nested_descriptor_expressions.add(call_expr)
                    self._nested_descriptor_expressions[arg] = type_resolver

    def _base_workflow_class_hook(self, ctx: ClassDefContext) -> None:
        """
        Placeholder for any special type logic we want to add to the Workflow class.
        """

        pass

    def get_function_signature_hook(self, fullname: str) -> Optional[Callable[[FunctionSigContext], FunctionLike]]:
        """
        This hook is used whenever we're calling a function and are type checking the signature. e.g. `f(a, b)`.

        We use this to support nested objects that reference descriptors within a node. Class initialization
        counts as a function call in mypy, so we want to support nested descriptors assigned to class instances
        we don't control, like dataclasses.
        """

        return self._function_signature_hook

    def _function_signature_hook(self, ctx: FunctionSigContext) -> FunctionLike:
        if not isinstance(ctx.context, CallExpr):
            return ctx.default_signature

        if ctx.context not in self._calls_with_nested_descriptor_expressions:
            return ctx.default_signature

        old_arg_types = ctx.default_signature.arg_types
        old_arg_names = ctx.default_signature.arg_names
        old_arg_kinds = ctx.default_signature.arg_kinds
        new_arg_types = []

        new_arg_by_name = {
            arg_name: ctx.context.args[arg_index] for arg_index, arg_name in enumerate(ctx.context.arg_names)
        }

        should_copy_new_signature = False

        for arg_index, old_arg_type in enumerate(old_arg_types):
            if arg_index >= len(old_arg_kinds) or arg_index >= len(old_arg_names):
                new_arg_types.append(old_arg_type)
                continue

            old_arg_kind = old_arg_kinds[arg_index]
            if old_arg_kind.is_named():
                old_arg_name = old_arg_names[arg_index]
                old_arg = new_arg_by_name.get(old_arg_name)
            elif arg_index < len(ctx.context.args):
                old_arg = ctx.context.args[arg_index]
            else:
                old_arg = None

            if isinstance(old_arg, MemberExpr) and old_arg in self._nested_descriptor_expressions:
                should_copy_new_signature = True

                new_arg_types.append(
                    self._get_resolvable_type(
                        self._nested_descriptor_expressions[old_arg],
                        old_arg_type,
                    )
                )
            else:
                new_arg_types.append(old_arg_type)

        if not should_copy_new_signature:
            return ctx.default_signature

        return ctx.default_signature.copy_modified(
            ret_type=ctx.default_signature.ret_type,
            arg_types=new_arg_types,
        )

    def _get_resolvable_type(
        self, get_named_type: Callable[[str, List[MypyType]], MypyType], type_: MypyType
    ) -> MypyType:
        if isinstance(type_, TypeAliasType) and type_.alias:
            if type_.alias.fullname == "vellum.workflows.types.core.Json":
                """
                We want to avoid infinite recursion, so we just state that the descriptor can
                just reference `Json` directly instead of each of the members.
                """
                return UnionType(
                    [
                        type_,
                        get_named_type("vellum.workflows.descriptors.base.BaseDescriptor", [type_]),
                    ]
                )

            """
            Type Aliases expand to an actual type, so we want to keep drilling down.
            Example: Foo = str
            """
            return self._get_resolvable_type(get_named_type, type_.alias.target)

        if isinstance(type_, UnionType):
            """
            If a node attribute is referencing a union type, we want to accept a descriptor
            pointing to any of the individual members.
            """

            return UnionType([self._get_resolvable_type(get_named_type, t) for t in type_.items])

        if isinstance(type_, Instance) and type_.type.fullname == "builtins.dict":
            """
            If a node attribute is referencing a dict, we want to accept a descriptor pointing
            to the dict itself or any of the values it maps to.
            """

            key_type = type_.args[0]
            value_type = type_.args[1]
            return get_named_type(
                type_.type.fullname,
                [
                    key_type,
                    self._get_resolvable_type(get_named_type, value_type),
                ],
            )

        """
        Otherwise by default, we want to accept a descriptor pointing to the type itself.
        """
        return UnionType(
            [
                type_,
                get_named_type("vellum.workflows.descriptors.base.BaseDescriptor", [type_]),
            ]
        )

    def get_attribute_hook(self, fullname: str) -> Optional[Callable[[AttributeContext], MypyType]]:
        return self._attribute_hook

    def _attribute_hook(self, ctx: AttributeContext) -> MypyType:
        if not isinstance(ctx.context, MemberExpr):
            return ctx.default_attr_type
        if not isinstance(ctx.context.expr, NameExpr) or ctx.context.expr.name != "self":
            return ctx.default_attr_type

        # TODO: ensure that `self` is a BaseNode
        # https://app.shortcut.com/vellum/story/5531

        return self._resolve_descriptor_type(ctx.default_attr_type)

    def _resolve_descriptor_type(self, default_type: MypyType) -> MypyType:
        if not isinstance(default_type, UnionType):
            return default_type

        non_descriptor_items = [
            (
                item.args[0]
                if isinstance(item, Instance)
                and _is_subclass(item.type, "vellum.workflows.descriptors.base.BaseDescriptor")
                and len(item.args) > 0
                else item
            )
            for item in default_type.items
        ]

        new_items = list(set(non_descriptor_items))

        if len(new_items) == 0:
            return default_type

        if len(new_items) == 1:
            return new_items[0]

        return UnionType(
            items=new_items,
        )

    def get_method_hook(self, fullname: str) -> Optional[Callable[[MethodContext], MypyType]]:
        """
        This hook is used whenever we're calling a method from a class and are type checking the return type.
        We use this to support special return types from our classes that isn't supported by default due to
        the lack of nested class inheritance, among other python pitfalls.
        """

        if fullname.endswith(".run"):
            return self._run_method_hook

        if fullname.endswith(".stream"):
            return self._stream_method_hook

        return None

    def _run_method_hook(self, ctx: MethodContext) -> MypyType:
        """
        We use this to target:
        - `BaseWorkflow.run()`, so that the WorkflowExecutionFulfilledEvent is properly typed
           using the `Outputs` class defined on the user-defined subclass of `Workflow`.
        - `BaseNode.run()`, so that the `Outputs` class defined on the user-defined subclass of `Node`
          is properly typed.
        """

        if isinstance(ctx.default_return_type, TypeAliasType):
            return self._workflow_run_method_hook(ctx)

        if isinstance(ctx.default_return_type, Instance):
            return self._node_run_method_hook(ctx)

        return ctx.default_return_type

    def _workflow_run_method_hook(self, ctx: MethodContext) -> MypyType:
        if not isinstance(ctx.default_return_type, TypeAliasType):
            return ctx.default_return_type

        alias = ctx.default_return_type.alias
        if not alias:
            return ctx.default_return_type

        alias_target = alias.target
        if not isinstance(alias_target, UnionType) or not alias_target.items:
            return ctx.default_return_type

        fulfilled_event = alias_target.items[0]
        if not isinstance(fulfilled_event, Instance):
            return ctx.default_return_type

        if fulfilled_event.type.fullname != "vellum.workflows.events.workflow.WorkflowExecutionFulfilledEvent":
            return ctx.default_return_type

        outputs_node = self._get_workflow_outputs_type_info(ctx)
        if not outputs_node:
            return ctx.default_return_type

        new_fulfilled_event = fulfilled_event.copy_modified(args=(Instance(outputs_node, []),))
        return TypeAliasType(
            alias=MypyTypeAlias(
                target=UnionType(
                    items=[new_fulfilled_event] + alias_target.items[1:],
                ),
                fullname=alias.fullname,
                line=alias.line,
                column=alias.column,
            ),
            args=ctx.default_return_type.args,
            line=ctx.default_return_type.line,
            column=ctx.default_return_type.column,
        )

    def _node_run_method_hook(self, ctx: MethodContext) -> MypyType:
        if not isinstance(ctx.default_return_type, Instance):
            return ctx.default_return_type

        if not _is_subclass(ctx.default_return_type.type, f"{BASE_NODE_FULLNAME}.Outputs"):
            return ctx.default_return_type

        outputs_node = self._get_node_outputs_type_info(ctx)
        if not outputs_node:
            return ctx.default_return_type

        return Instance(outputs_node, [])

    def _stream_method_hook(self, ctx: MethodContext) -> MypyType:
        """
        We use this to target `Workflow.stream()` so that the WorkflowExecutionFulfilledEvent is properly typed
        using the `Outputs` class defined on the user-defined subclass of `Workflow`.
        """

        if not isinstance(ctx.default_return_type, TypeAliasType):
            return ctx.default_return_type

        alias = ctx.default_return_type.alias
        if not alias:
            return ctx.default_return_type

        alias_target = alias.target
        if (
            not isinstance(alias_target, Instance)
            or not _is_subclass(alias_target.type, "vellum.workflows.events.stream.WorkflowEventGenerator")
            or not alias_target.args
        ):
            return ctx.default_return_type

        union_alias = alias_target.args[0]
        if not isinstance(union_alias, TypeAliasType) or not union_alias.alias:
            return ctx.default_return_type

        union_target = union_alias.alias.target
        if not isinstance(union_target, UnionType) or not union_target.items:
            return ctx.default_return_type

        fulfilled_event_index = -1
        fulfilled_event = None
        for event_type_index, event_type in enumerate(union_target.items):
            if not isinstance(event_type, Instance):
                continue

            if event_type.type.fullname != "vellum.workflows.events.workflow.WorkflowExecutionFulfilledEvent":
                continue

            fulfilled_event_index = event_type_index
            fulfilled_event = event_type

        if fulfilled_event_index == -1 or not fulfilled_event:
            return ctx.default_return_type

        outputs_node = self._get_workflow_outputs_type_info(ctx)
        if not outputs_node:
            return ctx.default_return_type

        new_fulfilled_event = fulfilled_event.copy_modified(args=(Instance(outputs_node, []),))
        return TypeAliasType(
            alias=MypyTypeAlias(
                target=alias_target.copy_modified(
                    args=[
                        TypeAliasType(
                            alias=MypyTypeAlias(
                                target=UnionType(
                                    items=[
                                        new_fulfilled_event if index == fulfilled_event_index else item
                                        for index, item in enumerate(union_target.items)
                                    ],
                                ),
                                fullname=union_alias.alias.fullname,
                                line=union_alias.alias.line,
                                column=union_alias.alias.column,
                            ),
                            args=union_alias.args,
                            line=union_alias.line,
                            column=union_alias.column,
                        )
                    ]
                    + list(alias_target.args[1:])
                ),
                fullname=alias.fullname,
                line=alias.line,
                column=alias.column,
            ),
            args=ctx.default_return_type.args,
            line=ctx.default_return_type.line,
            column=ctx.default_return_type.column,
        )

    def _get_workflow_outputs_type_info(self, ctx: MethodContext) -> Optional[TypeInfo]:
        if not isinstance(ctx.context, CallExpr):
            return None

        if not isinstance(ctx.context.callee, MemberExpr):
            return None

        expr = ctx.context.callee.expr
        instance = ctx.api.get_expression_type(expr)
        if not isinstance(instance, Instance) or not _is_subclass(
            instance.type, "vellum.workflows.workflows.base.BaseWorkflow"
        ):
            return None

        outputs_node = instance.type.names.get("Outputs")

        if (
            not outputs_node
            or not isinstance(outputs_node.node, TypeInfo)
            or not _is_subclass(outputs_node.node, "vellum.workflows.outputs.base.BaseOutputs")
        ):
            return None

        resolved_outputs_node = self._resolve_descriptors_in_outputs(outputs_node)

        if not isinstance(resolved_outputs_node.node, TypeInfo):
            return None

        return resolved_outputs_node.node

    def _get_node_outputs_type_info(self, ctx: MethodContext) -> Optional[TypeInfo]:
        if not isinstance(ctx.context, CallExpr):
            return None

        if not isinstance(ctx.context.callee, MemberExpr):
            return None

        expr = ctx.context.callee.expr
        instance = ctx.api.get_expression_type(expr)
        if not isinstance(instance, Instance) or not _is_subclass(instance.type, BASE_NODE_FULLNAME):
            return None

        outputs_node = instance.type.names.get("Outputs")

        if (
            not outputs_node
            or not isinstance(outputs_node.node, TypeInfo)
            or not _is_subclass(outputs_node.node, "vellum.workflows.outputs.base.BaseOutputs")
        ):
            return None

        # TODO: For some reason, returning the correct Outputs type info is causing `result` to not
        # be found. `test_templating_node.py` is the best place to test this.
        # https://app.shortcut.com/vellum/story/6132
        # return outputs_node.node
        return None

    def _resolve_descriptors_in_outputs(self, type_info: SymbolTableNode) -> SymbolTableNode:
        new_type_info = type_info.copy()
        if not isinstance(new_type_info.node, TypeInfo):
            return new_type_info

        for sym in new_type_info.node.names.values():
            if isinstance(sym.node, Var):
                descriptor_type = sym.node.type
                if isinstance(descriptor_type, Instance) and _is_subclass(
                    descriptor_type.type, "vellum.workflows.descriptors.base.BaseDescriptor"
                ):
                    args = descriptor_type.args
                    if args:
                        sym.node.type = args[0]

        return new_type_info


def plugin(version: str) -> Type[VellumMypyPlugin]:
    return VellumMypyPlugin



import pytest
import sys
from uuid import UUID, uuid4
from typing import Type, cast

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.client.types.code_executor_response import CodeExecutorResponse
from vellum.client.types.number_vellum_value import NumberVellumValue
from vellum.workflows import BaseWorkflow
from vellum.workflows.events.workflow import WorkflowExecutionInitiatedBody
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.nodes import BaseNode
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.utils.uuids import generate_workflow_deployment_prefix
from vellum.workflows.utils.zip import zip_file_map
from vellum_ee.workflows.display.utils.events import event_enricher
from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder


@pytest.fixture
def virtual_file_loader():
    """Fixture to manage VirtualFileFinder registration and cleanup."""
    finders = []

    def _register(files, namespace, source_module=None):
        finder = VirtualFileFinder(files, namespace, source_module=source_module)
        sys.meta_path.append(finder)
        finders.append(finder)
        return finder

    yield _register

    for finder in reversed(finders):
        if finder in sys.meta_path:
            sys.meta_path.remove(finder)


def test_load_workflow_event_display_context():
    from vellum.workflows.events.workflow import WorkflowEventDisplayContext

    # We are actually just ensuring there are no circular dependencies when
    # our Workflow Server imports this class.
    assert issubclass(WorkflowEventDisplayContext, UniversalBaseModel)


def test_load_from_module__lazy_reference_in_file_loader():
    # GIVEN a workflow module with a node containing a lazy reference
    files = {
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.start_node import StartNode

class Workflow(BaseWorkflow):
    graph = StartNode
""",
        "nodes/start_node.py": """\
from vellum.workflows.nodes import BaseNode
from vellum.workflows.references import LazyReference

class StartNode(BaseNode):
    foo = LazyReference(lambda: StartNode.Outputs.bar)

    class Outputs(BaseNode.Outputs):
        bar = str
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    sys.meta_path.append(VirtualFileFinder(files, namespace))

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow()

    # THEN the workflow is successfully initialized
    assert workflow

    # AND the graph is just a BaseNode
    # ideally this would be true, but the loader uses a different BaseNode class definition than
    # the one in this test module.
    # assert isinstance(workflow.graph, BaseNode)
    start_node = cast(Type[BaseNode], workflow.graph)
    assert start_node.__bases__ == (BaseNode,)

    # AND the lazy reference has the correct name
    assert start_node.foo.instance
    assert start_node.foo.instance.name == "StartNode.Outputs.bar"


def test_load_from_module__ts_code_in_file_loader(
    vellum_client,
):
    # GIVEN typescript code
    ts_code = """async function main(): any {
  return 5;
}"""

    # AND a workflow module with only a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from .nodes.code_execution_node import CodeExecutionNode

class Workflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
       final_output = CodeExecutionNode.Outputs.result
""",
        "nodes/code_execution_node/__init__.py": """
from typing import Any

from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

class CodeExecutionNode(BaseCodeExecutionNode[BaseState, int]):
    filepath = "./script.ts"
    code_inputs = {}
    runtime = "TYPESCRIPT_5_3_3"
    packages = []
""",
        "nodes/code_execution_node/script.ts": ts_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=5),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    event = workflow.run()
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result
    assert event.body.outputs == {"final_output": 5.0}

    # AND the workflow definition module is correctly serialized as a list
    serialized_event = event.model_dump(mode="json")
    workflow_definition = serialized_event["body"]["workflow_definition"]
    assert workflow_definition["module"] == [namespace, "workflow"]


def test_load_from_module__simple_code_execution_node_with_try(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with only a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.code_execution_node import CodeExecutionNode

class Workflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
       final_output = CodeExecutionNode.Outputs.result
""",
        "nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.core import TryNode
from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

@TryNode.wrap()
class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/code_execution_node/script.py": py_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=1),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result
    assert event.body.outputs == {"final_output": 1.0}


def test_load_from_module__simple_code_execution_node_with_retry(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with only a code execution node wrapped with RetryNode
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.code_execution_node import CodeExecutionNode

class Workflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
       final_output = CodeExecutionNode.Outputs.result
""",
        "nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.core.retry_node.node import RetryNode
from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

@RetryNode.wrap(max_attempts=3)
class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/code_execution_node/script.py": py_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=1),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result
    assert event.body.outputs == {"final_output": 1.0}


def test_load_from_module__code_execution_within_subworkflow(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with a subworkflow containing a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.subworkflow import Subworkflow

class Workflow(BaseWorkflow):
    graph = Subworkflow

    class Outputs(BaseWorkflow.Outputs):
       final_output = Subworkflow.Outputs.result
""",
        "nodes/subworkflow/__init__.py": """
from vellum.workflows.nodes.displayable import InlineSubworkflowNode

from .workflow import SubworkflowWorkflow

class Subworkflow(InlineSubworkflowNode):
    subworkflow = SubworkflowWorkflow
""",
        "nodes/subworkflow/nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/subworkflow/nodes/code_execution_node/script.py": py_code,
        "nodes/subworkflow/workflow.py": """
from vellum.workflows import BaseWorkflow

from .nodes.code_execution_node import CodeExecutionNode

class SubworkflowWorkflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
        result = CodeExecutionNode.Outputs.result
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=1),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result from the subworkflow
    assert event.body.outputs == {"final_output": 1.0}


def test_load_from_module__code_execution_within_map_node(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with a map node containing a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.map_node import MapNode

class Workflow(BaseWorkflow):
    graph = MapNode
    
    class Outputs(BaseWorkflow.Outputs):  # noqa: W293
       results = MapNode.Outputs.final_output
""",
        "nodes/map_node/__init__.py": """
from vellum.workflows.nodes.core.map_node import MapNode as BaseMapNode

from .workflow import MapNodeWorkflow

class MapNode(BaseMapNode):
    items = ["foo", "bar", "baz"]
    subworkflow = MapNodeWorkflow
    max_concurrency = 4
""",
        "nodes/map_node/nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/map_node/nodes/code_execution_node/script.py": py_code,
        "nodes/map_node/workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.code_execution_node import CodeExecutionNode

class MapNodeInputs:
    item: str
    index: int

class MapNodeWorkflow(BaseWorkflow):
    graph = CodeExecutionNode
    
    class Outputs(BaseWorkflow.Outputs):  # noqa: W293
        final_output = CodeExecutionNode.Outputs.result
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=3),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the map node results as a list
    assert event.body.outputs == {"results": [1.0, 1.0, 1.0]}


def test_load_from_module__syntax_error_in_node_file():
    """
    Tests that a syntax error in a node file raises WorkflowInitializationException with user-facing message.
    """
    # GIVEN a workflow module with a node file containing a syntax error (missing colon)
    files = {
        "__init__.py": "",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.broken_node import BrokenNode

class Workflow(BaseWorkflow):
    graph = BrokenNode
""",
        "nodes/broken_node.py": """\
from vellum.workflows.nodes import BaseNode

class BrokenNode(BaseNode)  # Missing colon
    \"\"\"This node has a syntax error.\"\"\"
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    sys.meta_path.append(VirtualFileFinder(files, namespace))

    # WHEN we attempt to load the workflow
    # THEN it should raise WorkflowInitializationException
    with pytest.raises(WorkflowInitializationException) as exc_info:
        BaseWorkflow.load_from_module(namespace)

    # AND the error message should be user-friendly
    error_message = str(exc_info.value)
    assert "Syntax Error raised while loading Workflow:" in error_message
    assert "invalid syntax" in error_message or "expected ':'" in error_message


def test_load_from_module__name_error_in_node_file():
    """
    Tests that a NameError in a node file raises WorkflowInitializationException with user-facing message.
    """
    # GIVEN a workflow module with a node file containing a NameError (undefined class reference)
    files = {
        "__init__.py": "",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.broken_node import BrokenNode

class Workflow(BaseWorkflow):
    graph = BrokenNode
""",
        "nodes/broken_node.py": """\
from vellum.workflows.nodes import BaseNode

class BrokenNode(BaseNode):
    some_attribute = UndefinedClass()
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    sys.meta_path.append(VirtualFileFinder(files, namespace))

    # WHEN we attempt to load the workflow
    # THEN it should raise WorkflowInitializationException
    with pytest.raises(WorkflowInitializationException) as exc_info:
        BaseWorkflow.load_from_module(namespace)

    # AND the error message should be user-friendly
    error_message = str(exc_info.value)
    assert "Invalid variable reference:" in error_message
    assert "UndefinedClass" in error_message or "not defined" in error_message


def test_load_from_module__module_not_found_error():
    """
    Tests that a ModuleNotFoundError raises WorkflowInitializationException with user-facing message.
    """
    # GIVEN a workflow module that imports a non-existent module
    files = {
        "__init__.py": "",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .non_existent_module import SomeClass

class Workflow(BaseWorkflow):
    graph = None
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    sys.meta_path.append(VirtualFileFinder(files, namespace, source_module="test"))

    # WHEN we attempt to load the workflow
    # THEN it should raise WorkflowInitializationException
    with pytest.raises(WorkflowInitializationException) as exc_info:
        BaseWorkflow.load_from_module(namespace)

    # AND the error message should be user-friendly and show source_module instead of namespace
    error_message = str(exc_info.value)
    assert error_message == "Workflow module not found: No module named 'test.non_existent_module'"


def test_load_from_module__module_not_found_error_with_external_package():
    """
    Tests that when ModuleNotFoundError occurs for an external package (not containing the namespace),
    the exception includes vellum_on_error_action set to CREATE_CUSTOM_IMAGE in raw_data.
    """

    # GIVEN a workflow module that imports a non-existent external package
    files = {
        "__init__.py": "",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
import some_external_package

class Workflow(BaseWorkflow):
    graph = None
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace, source_module="test")
    sys.meta_path.append(finder)

    # WHEN we attempt to load the workflow
    # THEN it should raise WorkflowInitializationException
    with pytest.raises(WorkflowInitializationException) as exc_info:
        BaseWorkflow.load_from_module(namespace)

    # AND the error message should be user-friendly
    error_message = str(exc_info.value)
    assert "Workflow module not found:" in error_message
    assert "some_external_package" in error_message

    assert exc_info.value.raw_data is not None
    assert exc_info.value.raw_data["vellum_on_error_action"] == "CREATE_CUSTOM_IMAGE"


def test_load_from_module__module_not_found_error_with_internal_package():
    """
    Tests that when ModuleNotFoundError occurs for an internal module (containing the namespace),
    the exception does NOT include vellum_on_error_action in raw_data.
    """

    # GIVEN a workflow module that imports a non-existent internal module
    files = {
        "__init__.py": "",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .non_existent_module import SomeClass

class Workflow(BaseWorkflow):
    graph = None
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace, source_module="test")
    sys.meta_path.append(finder)

    # WHEN we attempt to load the workflow
    # THEN it should raise WorkflowInitializationException
    with pytest.raises(WorkflowInitializationException) as exc_info:
        BaseWorkflow.load_from_module(namespace)

    # AND the error message should be user-friendly
    error_message = str(exc_info.value)
    assert "Workflow module not found:" in error_message

    assert exc_info.value.raw_data is None


def test_serialize_module__tool_calling_node_with_single_tool():
    """Test that serialize_module works with a tool calling node that has a single tool."""

    # GIVEN a simple tool function
    tool_function_code = '''def get_weather(location: str) -> str:
    """Get the current weather for a location."""
    return f"The weather in {location} is sunny."
'''

    # AND a workflow module with a tool calling node using that single tool
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes.displayable.tool_calling_node import ToolCallingNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.workflows.base import BaseInputs
from vellum.client.types.chat_message_prompt_block import ChatMessagePromptBlock
from vellum.client.types.plain_text_prompt_block import PlainTextPromptBlock
from vellum.client.types.rich_text_prompt_block import RichTextPromptBlock
from vellum.client.types.variable_prompt_block import VariablePromptBlock

from .get_weather import get_weather


class Inputs(BaseInputs):
    location: str


class WeatherNode(ToolCallingNode):
    ml_model = "gpt-4o-mini"
    blocks = [
        ChatMessagePromptBlock(
            chat_role="USER",
            blocks=[
                RichTextPromptBlock(
                    blocks=[
                        VariablePromptBlock(
                            input_variable="location",
                        ),
                    ],
                ),
            ],
        ),
    ]
    functions = [get_weather]
    prompt_inputs = {
        "location": Inputs.location,
    }


class Workflow(BaseWorkflow[Inputs, BaseState]):
    graph = WeatherNode

    class Outputs(BaseWorkflow.Outputs):
        result = WeatherNode.Outputs.text
""",
        "get_weather.py": tool_function_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    sys.meta_path.append(VirtualFileFinder(files, namespace))

    result = BaseWorkflowDisplay.serialize_module(namespace)

    # THEN the serialization should complete successfully
    assert result is not None


def test_resolve_workflow_deployment__returns_workflow_with_generated_files():
    """
    Test that resolve_workflow_deployment returns a workflow with artifacts
    in generated_files using the correct prefix format.
    """
    # GIVEN a deployment name and release tag
    deployment_name = "test_deployment"
    release_tag = "LATEST"

    expected_prefix = generate_workflow_deployment_prefix(deployment_name, release_tag)

    # Create a simple test node for the resolved workflow
    test_node_code = """
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.outputs import BaseOutputs

class TestNode(BaseNode):
    template = "Hello"

    class Outputs(BaseOutputs):
        result: str

    def run(self):
        return self.Outputs(result="Hello, {template}")
"""

    mock_workflow_code = """
from vellum.workflows import BaseWorkflow
from .nodes.test_node import TestNode

class ResolvedWorkflow(BaseWorkflow):
    graph = TestNode
"""

    # Create parent workflow files that reference the subworkflow deployment
    parent_workflow_code = """
from vellum.workflows import BaseWorkflow
from .nodes.subworkflow_deployment_node import TestSubworkflowDeploymentNode

class ParentWorkflow(BaseWorkflow):
    graph = TestSubworkflowDeploymentNode
"""

    parent_node_code = """
from vellum.workflows.nodes import SubworkflowDeploymentNode
from vellum.workflows.outputs import BaseOutputs

class TestSubworkflowDeploymentNode(SubworkflowDeploymentNode):
    deployment = "test_deployment"

    class Outputs(BaseOutputs):
        result: str

    subworkflow_inputs = {"message": "test"}
"""

    files = {
        "__init__.py": "",
        "workflow.py": parent_workflow_code,
        "nodes/subworkflow_deployment_node.py": parent_node_code,
        f"{expected_prefix}/__init__.py": "",
        f"{expected_prefix}/workflow.py": mock_workflow_code,
        f"{expected_prefix}/nodes/test_node.py": test_node_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # WHEN we execute the root workflow
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files, namespace=namespace))

    # THEN the workflow should be successfully initialized
    assert workflow

    event = workflow.run()

    # AND the method should return a workflow (not None) - this will pass once implemented
    assert event.name == "workflow.execution.fulfilled"


def test_resolve_workflow_deployment__uses_pull_api_with_inputs_deployment_name(vellum_client):
    """
    Test that resolve_workflow_deployment uses the pull API to fetch subworkflow files
    when the deployment name comes from Inputs.deployment_name.
    """
    # GIVEN a deployment name and release tag
    deployment_name = "test_deployment"
    release_tag = "LATEST"

    test_node_code = """
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.outputs import BaseOutputs

class TestNode(BaseNode):
    template = "Hello"

    class Outputs(BaseOutputs):
        result: str

    def run(self):
        return self.Outputs(result="Hello, {template}")
"""

    mock_workflow_code = """
from vellum.workflows import BaseWorkflow
from .nodes.test_node import TestNode

class ResolvedWorkflow(BaseWorkflow):
    graph = TestNode
"""

    inputs_code = """
from vellum.workflows.inputs import BaseInputs

class Inputs(BaseInputs):
    deployment_name: str
"""

    parent_workflow_code = """
from vellum.workflows import BaseWorkflow
from .inputs import Inputs
from .nodes.subworkflow_deployment_node import TestSubworkflowDeploymentNode
from vellum.workflows.state import BaseState

class ParentWorkflow(BaseWorkflow[Inputs, BaseState]):
    graph = TestSubworkflowDeploymentNode
"""

    parent_node_code = """
from vellum.workflows.nodes import SubworkflowDeploymentNode
from vellum.workflows.outputs import BaseOutputs
from ..inputs import Inputs

class TestSubworkflowDeploymentNode(SubworkflowDeploymentNode):
    deployment = Inputs.deployment_name

    class Outputs(BaseOutputs):
        result: str

    subworkflow_inputs = {"message": "test"}
"""

    subworkflow_files = {
        "__init__.py": "",
        "workflow.py": mock_workflow_code,
        "nodes/test_node.py": test_node_code,
    }

    parent_files = {
        "__init__.py": "",
        "inputs.py": inputs_code,
        "workflow.py": parent_workflow_code,
        "nodes/subworkflow_deployment_node.py": parent_node_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered for the parent workflow
    finder = VirtualFileFinder(parent_files, namespace)
    sys.meta_path.append(finder)

    vellum_client.workflows.pull.return_value = iter([zip_file_map(subworkflow_files)])

    # WHEN we execute the root workflow with the mocked client
    Workflow = BaseWorkflow.load_from_module(namespace)
    Inputs = Workflow.get_inputs_class()

    workflow = Workflow(context=WorkflowContext(namespace=namespace, generated_files=parent_files))
    final_event = workflow.run(inputs=Inputs(deployment_name=deployment_name))

    # THEN the method should return a workflow (not None) - this will pass once implemented
    assert final_event.name == "workflow.execution.fulfilled", final_event

    # AND the pull API should have been called with the correct deployment name, release tag, and version
    args, kwargs = vellum_client.workflows.pull.call_args
    assert args[0] == deployment_name
    assert kwargs["release_tag"] == release_tag
    assert kwargs["version"].startswith(">=")
    assert ".0.0,<=" in kwargs["version"]

    # AND the X-Vellum-Always-Success header should be included for graceful error handling
    assert kwargs["request_options"]["additional_headers"]["X-Vellum-Always-Success"] == "true"


def test_workflow_initiated_event_includes_display_context_with_output_display_name():
    """
    Tests that workflow initiated events include display context with annotated node and output information.
    """
    # GIVEN a workflow with a single node that has a single output
    files = {
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.start_node import StartNode

class Workflow(BaseWorkflow):
    graph = StartNode

    class Outputs(BaseWorkflow.Outputs):
        final_result = StartNode.Outputs.result
""",
        "nodes/start_node.py": """\
from vellum.workflows.nodes import BaseNode

class StartNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        result: str = "test output"
""",
        "display/nodes/__init__.py": """\
from .start_node import StartNodeDisplay
""",
        "display/nodes/start_node.py": """\
from uuid import UUID
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay
from ...nodes.start_node import StartNode

class StartNodeDisplay(BaseNodeDisplay[StartNode]):
    node_id = UUID("11111111-1111-1111-1111-111111111111")
    output_display = {
        StartNode.Outputs.result: NodeOutputDisplay(
            id=UUID("22222222-2222-2222-2222-222222222222"),
            name="Pretty Result"
        )
    }
""",
        "display/workflow.py": """\
""",
    }

    namespace = str(uuid4())
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    try:
        # WHEN we stream the workflow and enrich the initiated event with display context
        Workflow = BaseWorkflow.load_from_module(namespace)
        workflow = Workflow()
        initiated_event = list(workflow.stream())[0]
        enriched_event = event_enricher(initiated_event)

        # THEN the initiated event should have display context
        assert isinstance(enriched_event.body, WorkflowExecutionInitiatedBody)
        body = enriched_event.body
        assert body.display_context is not None

        # AND the display context should contain the annotated node
        node_displays = body.display_context.node_displays
        annotated_node_id = UUID("11111111-1111-1111-1111-111111111111")
        assert annotated_node_id in node_displays

        # AND the node's output display should contain the output with the display name "Pretty Result"
        node_display = node_displays[annotated_node_id]
        assert "result" in node_display.output_display

        # AND the output should map to the annotated output id
        annotated_output_id = UUID("22222222-2222-2222-2222-222222222222")
        assert node_display.output_display["result"] == annotated_output_id
    finally:
        if finder in sys.meta_path:
            sys.meta_path.remove(finder)


def test_load_from_module__chained_nodes_with_absolute_imports(virtual_file_loader):
    """
    Tests that a workflow with chained nodes using absolute imports works correctly.
    This test is expected to fail on main as absolute imports are not yet supported.
    """

    # GIVEN a workflow module with two chained base nodes using absolute imports
    namespace = str(uuid4())
    source_module = "test_workflow_abs"
    files = {
        "__init__.py": "",
        "workflow.py": f"""
from vellum.workflows import BaseWorkflow
from {source_module}.nodes.first_node import FirstNode
from {source_module}.nodes.second_node import SecondNode

class Workflow(BaseWorkflow):
    graph = FirstNode >> SecondNode

    class Outputs(BaseWorkflow.Outputs):
        final_output = SecondNode.Outputs.result
""",
        "nodes/__init__.py": "",
        "nodes/first_node.py": """
from vellum.workflows.nodes import BaseNode

class FirstNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value: str

    def run(self) -> "FirstNode.Outputs":
        return self.Outputs(value="Hello from FirstNode")
""",
        "nodes/second_node.py": f"""
from vellum.workflows.nodes import BaseNode
from {source_module}.nodes.first_node import FirstNode

class SecondNode(BaseNode):
    input_value = FirstNode.Outputs.value

    class Outputs(BaseNode.Outputs):
        result: str

    def run(self) -> "SecondNode.Outputs":
        return self.Outputs(result=f"{{self.input_value}} -> SecondNode")
""",
    }

    # AND the virtual file loader is registered
    virtual_file_loader(files, namespace, source_module=source_module)

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow()

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    terminal_event = workflow.run()

    # THEN the workflow should have completed successfully
    assert terminal_event.name == "workflow.execution.fulfilled"

    # AND the outputs should be as expected
    assert terminal_event.outputs == {
        "final_output": "Hello from FirstNode -> SecondNode",
    }



import pytest
from uuid import UUID

from vellum.workflows.events.types import NodeParentContext, WorkflowDeploymentParentContext, WorkflowParentContext
from vellum.workflows.types.definition import CodeResourceDefinition


@pytest.fixture
def mock_complex_parent_context():
    # TODO: We were able to confirm that this parent context caused our serialization to hang, but don't know why yet.
    # We should try to reduce this example further to isolate a minimal example that reproduces the issue.
    return NodeParentContext(
        span_id=UUID("d697f8c8-b363-4154-8469-eb4f9fb5e445"),
        parent=WorkflowParentContext(
            span_id=UUID("a0c68884-22c3-4ac9-8476-8747884d80e1"),
            parent=NodeParentContext(
                span_id=UUID("46163407-71f7-40f2-9f66-872d4b338fcc"),
                parent=WorkflowParentContext(
                    span_id=UUID("0ddf01e7-d0c3-426c-af27-d8bfb22fcdd5"),
                    parent=NodeParentContext(
                        span_id=UUID("79a6c926-b5f3-4ede-b2f8-4bb6f0c086ba"),
                        parent=WorkflowParentContext(
                            span_id=UUID("530a56fe-90fd-4f4c-b905-457975fb3e10"),
                            parent=WorkflowDeploymentParentContext(
                                span_id=UUID("3e10a8c2-558c-4ef7-926d-7b79ebc7cba9"),
                                parent=NodeParentContext(
                                    span_id=UUID("a3cd4086-c0b9-4dff-88f3-3e2191b8a2a7"),
                                    parent=WorkflowParentContext(
                                        span_id=UUID("c2ba7577-8d24-49b1-aa92-b9ace8244090"),
                                        workflow_definition=CodeResourceDefinition(
                                            id=UUID("2e2d5c56-49b7-48b5-82fa-e80e72768b9c"),
                                            name="Workflow",
                                            module=["e81a6124-2c57-4c39-938c-ab6059059ff2", "workflow"],
                                        ),
                                    ),
                                    node_definition=CodeResourceDefinition(
                                        id=UUID("23d25675-f377-4450-916f-39ebee5c8ea9"),
                                        name="SubworkflowDeployment",
                                        module=[
                                            "e81a6124-2c57-4c39-938c-ab6059059ff2",
                                            "nodes",
                                            "subworkflow_deployment",
                                        ],
                                    ),
                                ),
                                deployment_id=UUID("cfc99610-2869-4506-b106-3fd7ce0bbb15"),
                                deployment_name="my-deployment",
                                deployment_history_item_id=UUID("13f31aae-29fd-4066-a4ec-c7687faebae3"),
                                release_tag_id=UUID("2d03987a-dcb5-49b9-981e-5e871c8f5d97"),
                                release_tag_name="LATEST",
                                external_id=None,
                                metadata=None,
                                workflow_version_id=UUID("7eaae816-b5f3-436d-8597-e8c3e4a32958"),
                            ),
                            workflow_definition=CodeResourceDefinition(
                                id=UUID("2e2d5c56-49b7-48b5-82fa-e80e72768b9c"),
                                name="Workflow",
                                module=["3e10a8c2-558c-4ef7-926d-7b79ebc7cba9", "workflow"],
                            ),
                        ),
                        node_definition=CodeResourceDefinition(
                            id=UUID("42c8adc2-a0d6-499e-81a4-e2e02d7beba9"),
                            name="MyNode",
                            module=[
                                "3e10a8c2-558c-4ef7-926d-7b79ebc7cba9",
                                "nodes",
                                "my_node",
                            ],
                        ),
                    ),
                    workflow_definition=CodeResourceDefinition(
                        id=UUID("b8563da0-7fd4-42e0-a75e-9ef037fca5a1"),
                        name="MyNodeWorkflow",
                        module=[
                            "3e10a8c2-558c-4ef7-926d-7b79ebc7cba9",
                            "nodes",
                            "my_node",
                            "workflow",
                        ],
                    ),
                ),
                node_definition=CodeResourceDefinition(
                    id=UUID("d44aee53-3b6e-41fd-8b7a-908cb2c77821"),
                    name="RetryNode",
                    module=[
                        "3e10a8c2-558c-4ef7-926d-7b79ebc7cba9",
                        "nodes",
                        "my_node",
                        "nodes",
                        "my_prompt",
                        "MyPrompt",
                        "<adornment>",
                    ],
                ),
            ),
            workflow_definition=CodeResourceDefinition(
                id=UUID("568a28dd-7134-436e-a5f4-790675212b51"),
                name="Subworkflow",
                module=["vellum", "workflows", "nodes", "utils"],
            ),
        ),
        node_definition=CodeResourceDefinition(
            id=UUID("86a34e5c-2652-49f0-9f9e-c653cf70029a"),
            name="MyPrompt",
            module=[
                "3e10a8c2-558c-4ef7-926d-7b79ebc7cba9",
                "nodes",
                "my_node",
                "nodes",
                "my_prompt",
            ],
        ),
    )



from dataclasses import dataclass
from uuid import UUID
from typing import Any, ClassVar, Dict, Generic, List, Optional, Tuple, TypeVar, Union, cast

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.expressions.and_ import AndExpression
from vellum.workflows.expressions.between import BetweenExpression
from vellum.workflows.expressions.is_nil import IsNilExpression
from vellum.workflows.expressions.is_not_nil import IsNotNilExpression
from vellum.workflows.expressions.is_not_null import IsNotNullExpression
from vellum.workflows.expressions.is_not_undefined import IsNotUndefinedExpression
from vellum.workflows.expressions.is_null import IsNullExpression
from vellum.workflows.expressions.is_undefined import IsUndefinedExpression
from vellum.workflows.expressions.not_between import NotBetweenExpression
from vellum.workflows.expressions.or_ import OrExpression
from vellum.workflows.nodes.displayable import ConditionalNode
from vellum.workflows.types.core import ConditionType, JsonObject
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.expressions import convert_descriptor_to_operator
from vellum_ee.workflows.display.vellum import NodeInput

_ConditionalNodeType = TypeVar("_ConditionalNodeType", bound=ConditionalNode)


@dataclass
class RuleIdMap:
    id: str
    lhs: Optional["RuleIdMap"]
    rhs: Optional["RuleIdMap"]
    field_node_input_id: Optional[str]
    value_node_input_id: Optional[str]


@dataclass
class ConditionId:
    id: str
    rule_group_id: Optional[str]


class BaseConditionalNodeDisplay(BaseNodeDisplay[_ConditionalNodeType], Generic[_ConditionalNodeType]):
    source_handle_ids: ClassVar[Dict[int, UUID]]
    rule_ids: ClassVar[List[RuleIdMap]]
    condition_ids: ClassVar[list[ConditionId]]

    def serialize(self, display_context: WorkflowDisplayContext, **_kwargs) -> JsonObject:
        node = self._node
        node_id = self.node_id

        node_inputs: List[NodeInput] = []
        source_handle_ids = self._get_source_handle_ids() or {}
        condition_ids = self._get_condition_ids()

        ports_size = sum(1 for _ in node.Ports)

        if len(condition_ids) > ports_size:
            raise ValueError(
                f"""Too many defined condition ids. Ports are size {ports_size} \
but the defined conditions have length {len(condition_ids)}"""
            )

        ports = self.serialize_ports(display_context)
        ports_by_index = {idx: port for idx, port in enumerate(ports)}

        def serialize_rule(
            descriptor: BaseDescriptor, path: List[int], rule_id_map: Optional[RuleIdMap]
        ) -> Dict[str, Any]:
            rule_id = (
                str(rule_id_map.id)
                if rule_id_map is not None
                else str(uuid4_from_hash(f"{node_id}|rule|{','.join(str(p) for p in path)}"))
            )

            result = self.get_nested_rule_details_by_path(rule_ids, path) if rule_ids else None
            if result is None:
                result = self._generate_hash_for_rule_ids(node_id, rule_id)
            current_id, field_node_input_id, value_node_input_id = result

            # Recursive step: Keep recursing until we hit the other descriptors
            if isinstance(descriptor, (AndExpression, OrExpression)):
                combinator = "AND" if isinstance(descriptor, AndExpression) else "OR"

                lhs = serialize_rule(descriptor._lhs, path + [0], rule_id_map.lhs if rule_id_map else None)
                rhs = serialize_rule(descriptor._rhs, path + [1], rule_id_map.rhs if rule_id_map else None)
                rules = [lhs, rhs]

                return {
                    "id": rule_id,
                    "rules": rules,
                    "combinator": combinator,
                    "negated": False,
                    "field_node_input_id": None,
                    "operator": None,
                    "value_node_input_id": None,
                }

            # Base cases for other descriptors
            elif isinstance(
                descriptor,
                (
                    IsNullExpression,
                    IsNotNullExpression,
                    IsNilExpression,
                    IsNotNilExpression,
                    IsUndefinedExpression,
                    IsNotUndefinedExpression,
                ),
            ):
                expression_node_input = create_node_input(
                    node_id, f"{current_id}.field", descriptor._expression, display_context, field_node_input_id
                )
                node_inputs.append(expression_node_input)
                field_node_input_id = expression_node_input.id
                operator = convert_descriptor_to_operator(descriptor)

            elif isinstance(descriptor, (BetweenExpression, NotBetweenExpression)):
                field_node_input = create_node_input(
                    node_id, f"{current_id}.field", descriptor._value, display_context, field_node_input_id
                )
                value_node_input = create_node_input(
                    node_id,
                    f"{current_id}.value",
                    f"{descriptor._start},{descriptor._end}",
                    display_context,
                    value_node_input_id,
                )
                node_inputs.extend([field_node_input, value_node_input])
                operator = convert_descriptor_to_operator(descriptor)
                field_node_input_id = field_node_input.id
                value_node_input_id = value_node_input.id

            else:
                lhs = descriptor._lhs  # type: ignore[attr-defined]
                rhs = descriptor._rhs  # type: ignore[attr-defined]

                lhs_node_input = create_node_input(
                    node_id, f"{current_id}.field", lhs, display_context, field_node_input_id
                )
                node_inputs.append(lhs_node_input)

                if descriptor._rhs is not None:  # type: ignore[attr-defined]
                    rhs_node_input = create_node_input(
                        node_id, f"{current_id}.value", rhs, display_context, value_node_input_id
                    )
                    node_inputs.append(rhs_node_input)
                    value_node_input_id = rhs_node_input.id

                operator = convert_descriptor_to_operator(descriptor)
                field_node_input_id = lhs_node_input.id

            return {
                "id": current_id,
                "rules": None,
                "combinator": None,
                "negated": False,
                "field_node_input_id": field_node_input_id,
                "operator": operator,
                "value_node_input_id": value_node_input_id,
            }

        conditions = []
        for idx, port in enumerate(node.Ports):

            condition_id = str(
                condition_ids[idx].id if condition_ids else uuid4_from_hash(f"{node_id}|conditions|{idx}")
            )
            rule_group_id = str(
                condition_ids[idx].rule_group_id if condition_ids else uuid4_from_hash(f"{condition_id}|rule_group")
            )
            source_handle_id = self._get_source_handle_id(idx, source_handle_ids, ports_by_index)

            if port._condition is None:
                if port._condition_type == ConditionType.ELSE:
                    conditions.append(
                        {
                            "id": condition_id,
                            "type": ConditionType.ELSE.value,
                            "source_handle_id": source_handle_id,
                            "data": None,
                        }
                    )
                else:
                    continue

            else:
                rule_ids = self._get_rule_ids()
                condition_rule = serialize_rule(port._condition, [idx], rule_ids[idx] if len(rule_ids) > idx else None)
                rules = condition_rule["rules"] if condition_rule["rules"] else [condition_rule]
                if port._condition_type is None:
                    raise ValueError("Condition type is None, but a valid ConditionType is expected.")
                conditions.append(
                    {
                        "id": condition_id,
                        "type": port._condition_type.value,
                        "source_handle_id": source_handle_id,
                        "data": {
                            "id": rule_group_id,
                            "rules": rules,
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    }
                )

        return {
            "id": str(node_id),
            "type": "CONDITIONAL",
            "inputs": [node_input.dict() for node_input in node_inputs],
            "data": {
                "label": self.label,
                "target_handle_id": str(self.get_target_handle_id()),
                "conditions": conditions,  # type: ignore
                "version": "2",
            },
            **self.serialize_generic_fields(display_context, exclude=["outputs"]),
        }

    def get_nested_rule_details_by_path(
        self, rule_ids: List[RuleIdMap], path: List[int]
    ) -> Union[Tuple[str, Optional[str], Optional[str]], None]:
        current_rule = rule_ids[path[0]]

        for step in path[1:]:
            if step == 0 and current_rule.lhs:
                current_rule = current_rule.lhs
            elif step == 1 and current_rule.rhs:
                current_rule = current_rule.rhs
            else:
                return None

        # This is essentially a leaf
        if current_rule.lhs and current_rule.lhs.lhs is None and current_rule.lhs.rhs is None:
            return (
                current_rule.lhs.id,
                current_rule.lhs.field_node_input_id,
                current_rule.lhs.value_node_input_id,
            )

        return None

    def _generate_hash_for_rule_ids(self, node_id, rule_id) -> Tuple[str, str, str]:
        return (
            str(uuid4_from_hash(f"{node_id}|{rule_id}|current")),
            str(uuid4_from_hash(f"{node_id}|{rule_id}||field")),
            str(uuid4_from_hash(f"{node_id}|{rule_id}||value")),
        )

    def _get_source_handle_ids(self) -> Dict[int, UUID]:
        return self._get_explicit_node_display_attr("source_handle_ids", Dict[int, UUID]) or {}

    def _get_rule_ids(self) -> List[RuleIdMap]:
        return self._get_explicit_node_display_attr("rule_ids", List[RuleIdMap]) or []

    def _get_condition_ids(self) -> List[ConditionId]:
        return self._get_explicit_node_display_attr("condition_ids", List[ConditionId]) or []

    def _get_source_handle_id(
        self, idx: int, source_handle_ids: Dict[int, UUID], ports_by_index: Dict[int, Any]
    ) -> str:
        if source_handle_ids.get(idx):
            return str(source_handle_ids[idx])

        if idx in ports_by_index:
            return cast(str, ports_by_index[idx]["id"])

        return str(uuid4_from_hash(f"{self.node_id}|source_handle|{idx}"))



import pydantic

IS_PYDANTIC_V1 = pydantic.VERSION.startswith("1.")
_loaded = False


def load_runtime_plugins() -> None:
    if IS_PYDANTIC_V1:
        # Pydantic plugins are only available in v2, so we defer the imports
        # below until we confirm we are running a supported version of pydantic
        return

    from pydantic.plugin import _loader as _pydantic_plugin_loader

    from vellum.plugins.pydantic import pydantic_plugin

    global _loaded
    if _loaded:
        return
    _loaded = True

    # TODO: This is a hack to get the Vellum plugin to load. We're supposed to use
    # pyproject.toml, but I couldn't figure out after an hour
    # https://app.shortcut.com/vellum/story/4635
    _pydantic_plugin_loader.get_plugins()
    if _pydantic_plugin_loader._plugins is not None:
        _pydantic_plugin_loader._plugins["vellum_plugin"] = pydantic_plugin



from typing import Any, Generic, Protocol, TypeVar, Union, runtime_checkable
from typing_extensions import TypeGuard

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.descriptors.utils import resolve_value
from vellum.workflows.state.base import BaseState


@runtime_checkable
class SupportsLen(Protocol):
    def __len__(self) -> int: ...


def has_len(obj: Any) -> TypeGuard[SupportsLen]:
    return hasattr(obj, "__len__")


_T = TypeVar("_T")


class LengthExpression(BaseDescriptor[int], Generic[_T]):
    def __init__(
        self,
        *,
        expression: Union[BaseDescriptor[_T], _T],
    ) -> None:
        super().__init__(name=f"length({expression})", types=(int,))
        self._expression = expression

    def resolve(self, state: "BaseState") -> int:
        expression = resolve_value(self._expression, state)

        if expression is undefined:
            raise InvalidExpressionException("Cannot get length of undefined value")

        if not has_len(expression):
            raise InvalidExpressionException(
                f"Expected an object that supports `len()`, got `{expression.__class__.__name__}`"
            )

        try:
            return len(expression)
        except TypeError as e:
            raise InvalidExpressionException(f"Cannot get length of `{expression.__class__.__name__}`: {str(e)}")



# This file was auto-generated by Fern from our API Definition.

import typing

WorkflowExecutionUsageCalculationErrorCodeEnum = typing.Union[
    typing.Literal["UNKNOWN", "DEPENDENCIES_FAILED", "NO_USAGE_CALCULATED", "INTERNAL_SERVER_ERROR"], typing.Any
]



from uuid import uuid4
from typing import Generator

from vellum.workflows.context import get_execution_context
from vellum.workflows.events.stream import WorkflowEventGenerator
from vellum.workflows.events.workflow import (
    WorkflowEvent,
    WorkflowEventStream,
    WorkflowExecutionInitiatedBody,
    WorkflowExecutionInitiatedEvent,
    WorkflowExecutionRejectedBody,
    WorkflowExecutionRejectedEvent,
)
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.inputs import BaseInputs


def stream_initialization_exception(
    exception: WorkflowInitializationException,
) -> WorkflowEventStream:
    """
    Stream a workflow initiated event followed by a workflow rejected event for an initialization exception.

    Args:
        exception: The WorkflowInitializationException to stream events for

    Returns:
        WorkflowEventGenerator yielding initiated and rejected events
    """

    execution_context = get_execution_context()
    span_id = uuid4()

    def _generate_events() -> Generator[WorkflowEvent, None, None]:
        initiated_event: WorkflowEvent = WorkflowExecutionInitiatedEvent(
            trace_id=execution_context.trace_id,
            span_id=span_id,
            body=WorkflowExecutionInitiatedBody(
                workflow_definition=exception.definition,
                inputs=BaseInputs(),
                initial_state=None,
            ),
            parent=execution_context.parent_context,
        )
        yield initiated_event

        rejected_event = WorkflowExecutionRejectedEvent(
            trace_id=execution_context.trace_id,
            span_id=span_id,
            body=WorkflowExecutionRejectedBody(
                workflow_definition=exception.definition,
                error=exception.error,
            ),
            parent=execution_context.parent_context,
        )
        yield rejected_event

    return WorkflowEventGenerator(_generate_events(), span_id)



# This file was auto-generated by Fern from our API Definition.

import typing

WorkflowExecutionEventErrorCode = typing.Union[
    typing.Literal[
        "WORKFLOW_INITIALIZATION",
        "WORKFLOW_CANCELLED",
        "WORKFLOW_TIMEOUT",
        "PROVIDER_CREDENTIALS_UNAVAILABLE",
        "INTEGRATION_CREDENTIALS_UNAVAILABLE",
        "NODE_EXECUTION_COUNT_LIMIT_REACHED",
        "INTERNAL_SERVER_ERROR",
        "NODE_EXECUTION",
        "NODE_CANCELLED",
        "NODE_TIMEOUT",
        "LLM_PROVIDER",
        "INVALID_TEMPLATE",
        "INVALID_INPUTS",
        "PROVIDER_QUOTA_EXCEEDED",
        "USER_DEFINED_ERROR",
    ],
    typing.Any,
]



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .list_workflow_release_tags_request_source import ListWorkflowReleaseTagsRequestSource
from .workflow_deployments_list_request_status import WorkflowDeploymentsListRequestStatus

__all__ = ["ListWorkflowReleaseTagsRequestSource", "WorkflowDeploymentsListRequestStatus"]



import os
from typing import List, Optional

from vellum import Vellum, VellumEnvironment
from vellum.client.types.api_version_enum import ApiVersionEnum


def create_vellum_client(
    api_key: Optional[str] = None,
    api_url: Optional[str] = None,
    api_version: Optional[ApiVersionEnum] = None,
) -> Vellum:
    if api_key is None:
        api_key = os.getenv("VELLUM_API_KEY", default="")

    return Vellum(
        api_key=api_key,
        environment=create_vellum_environment(api_url),
        api_version=api_version,
    )


def create_vellum_environment(api_url: Optional[str] = None) -> VellumEnvironment:
    return VellumEnvironment(
        default=_resolve_env([api_url, "VELLUM_DEFAULT_API_URL", "VELLUM_API_URL"], "https://api.vellum.ai"),
        documents=_resolve_env([api_url, "VELLUM_DOCUMENTS_API_URL", "VELLUM_API_URL"], "https://documents.vellum.ai"),
        predict=_resolve_env([api_url, "VELLUM_PREDICT_API_URL", "VELLUM_API_URL"], "https://predict.vellum.ai"),
    )


def _resolve_env(names: List[Optional[str]], default: str = "") -> str:
    for name in names:
        if not name:
            continue

        value = os.getenv(name)
        if value:
            return value

    return default



from functools import cached_property
from queue import Queue
from uuid import UUID, uuid4
from typing import TYPE_CHECKING, Any, Generator, Generic, Optional, Tuple, Type, TypeVar, cast

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor

if TYPE_CHECKING:
    from vellum.workflows.outputs import BaseOutputs
    from vellum.workflows.state.base import BaseState


_OutputType = TypeVar("_OutputType")


class OutputReference(BaseDescriptor[_OutputType], Generic[_OutputType]):
    def __init__(
        self,
        *,
        name: str,
        types: Tuple[Type[_OutputType], ...],
        instance: Optional[_OutputType],
        outputs_class: Type["BaseOutputs"],
    ) -> None:
        super().__init__(name=name, types=types, instance=instance)
        self._outputs_class = outputs_class

    @property
    def outputs_class(self) -> Type["BaseOutputs"]:
        return self._outputs_class

    @cached_property
    def id(self) -> UUID:
        parent_class = self._outputs_class.__parent_class__
        if not parent_class:
            return uuid4()

        output_ids = getattr(parent_class, "__output_ids__", {})
        if not isinstance(output_ids, dict):
            return uuid4()

        output_id = output_ids.get(self.name)
        if not isinstance(output_id, UUID):
            return uuid4()

        return output_id

    def resolve(self, state: "BaseState") -> _OutputType:
        node_output = state.meta.node_outputs.get(self, undefined)
        if isinstance(node_output, Queue):
            # Fix typing surrounding the return value of node outputs
            # https://app.shortcut.com/vellum/story/4783
            return self._as_generator(node_output)  # type: ignore[return-value]

        if node_output is not undefined:
            return cast(_OutputType, node_output)

        if state.meta.parent:
            return self.resolve(state.meta.parent)

        # Fix typing surrounding the return value of node outputs
        # https://app.shortcut.com/vellum/story/4783
        return cast(Type[undefined], node_output)  # type: ignore[return-value]

    def _as_generator(self, node_output: Queue) -> Generator[_OutputType, None, Type[undefined]]:
        while True:
            item = node_output.get()
            if item is undefined:
                return undefined
            yield cast(_OutputType, item)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return super().__eq__(other) and self._outputs_class == other._outputs_class

    def __hash__(self) -> int:
        return hash((self._outputs_class, self._name))

    def __repr__(self) -> str:
        return f"{self._outputs_class.__qualname__}.{self.name}"

    def __deepcopy__(self, memo: dict) -> "OutputReference[_OutputType]":
        return OutputReference(
            name=self._name, types=self._types, instance=self._instance, outputs_class=self._outputs_class
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.jsonable_encoder import jsonable_encoder
from ...core.pydantic_utilities import parse_obj_as
from ...core.request_options import RequestOptions
from ...types.container_image_read import ContainerImageRead
from ...types.docker_service_token import DockerServiceToken
from ...types.paginated_container_image_read_list import PaginatedContainerImageReadList

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawContainerImagesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PaginatedContainerImageReadList]:
        """
        Retrieve a list of container images for the organization.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PaginatedContainerImageReadList]

        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/container-images",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
                "ordering": ordering,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedContainerImageReadList,
                    parse_obj_as(
                        type_=PaginatedContainerImageReadList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def retrieve(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[ContainerImageRead]:
        """
        Retrieve a Container Image by its ID or name.

        Parameters
        ----------
        id : str
            Either the Container Image's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ContainerImageRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/container-images/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ContainerImageRead,
                    parse_obj_as(
                        type_=ContainerImageRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def docker_service_token(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DockerServiceToken]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DockerServiceToken]

        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/container-images/docker-service-token",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DockerServiceToken,
                    parse_obj_as(
                        type_=DockerServiceToken,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def push_container_image(
        self,
        *,
        name: str,
        sha: str,
        tags: typing.Sequence[str],
        force: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[ContainerImageRead]:
        """
        Parameters
        ----------
        name : str

        sha : str

        tags : typing.Sequence[str]

        force : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ContainerImageRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/container-images/push",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "name": name,
                "sha": sha,
                "tags": tags,
                "force": force,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ContainerImageRead,
                    parse_obj_as(
                        type_=ContainerImageRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawContainerImagesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PaginatedContainerImageReadList]:
        """
        Retrieve a list of container images for the organization.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PaginatedContainerImageReadList]

        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/container-images",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
                "ordering": ordering,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedContainerImageReadList,
                    parse_obj_as(
                        type_=PaginatedContainerImageReadList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def retrieve(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[ContainerImageRead]:
        """
        Retrieve a Container Image by its ID or name.

        Parameters
        ----------
        id : str
            Either the Container Image's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ContainerImageRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/container-images/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ContainerImageRead,
                    parse_obj_as(
                        type_=ContainerImageRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def docker_service_token(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DockerServiceToken]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DockerServiceToken]

        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/container-images/docker-service-token",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DockerServiceToken,
                    parse_obj_as(
                        type_=DockerServiceToken,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def push_container_image(
        self,
        *,
        name: str,
        sha: str,
        tags: typing.Sequence[str],
        force: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[ContainerImageRead]:
        """
        Parameters
        ----------
        name : str

        sha : str

        tags : typing.Sequence[str]

        force : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ContainerImageRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/container-images/push",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "name": name,
                "sha": sha,
                "tags": tags,
                "force": force,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ContainerImageRead,
                    parse_obj_as(
                        type_=ContainerImageRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



import pytest

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.expressions.length import LengthExpression
from vellum.workflows.state.base import BaseState


class TestState(BaseState):
    string_value: str = "hello world"


def test_length_expression_string():
    """
    Tests that LengthExpression correctly returns the length of a string.
    """

    state = TestState()

    expression = TestState.string_value.length()
    result = expression.resolve(state)

    assert result == 11


def test_length_expression_undefined():
    """
    Tests that LengthExpression raises an exception for undefined values.
    """

    expression = LengthExpression(expression=undefined)
    state = TestState()

    # THEN we should get an InvalidExpressionException
    with pytest.raises(InvalidExpressionException) as exc_info:
        expression.resolve(state)

    assert "Cannot get length of undefined value" in str(exc_info.value)



import json
from uuid import UUID, uuid4
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Generic, Iterator, List, Optional, Set, Type, Union, cast

from vellum import (
    ChatMessage,
    WorkflowExpandMetaRequest,
    WorkflowOutput,
    WorkflowRequestChatHistoryInputRequest,
    WorkflowRequestInputRequest,
    WorkflowRequestJsonInputRequest,
    WorkflowRequestNumberInputRequest,
    WorkflowRequestStringInputRequest,
)
from vellum.client.core import RequestOptions
from vellum.client.core.api_error import ApiError
from vellum.client.types.chat_message_request import ChatMessageRequest
from vellum.workflows.constants import LATEST_RELEASE_TAG, OMIT, undefined
from vellum.workflows.context import execution_context, get_execution_context, get_parent_context
from vellum.workflows.errors import WorkflowErrorCode
from vellum.workflows.errors.types import workflow_event_error_to_workflow_error
from vellum.workflows.events.types import WorkflowDeploymentParentContext, default_serializer
from vellum.workflows.events.workflow import is_workflow_event
from vellum.workflows.exceptions import NodeException, WorkflowInitializationException
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.outputs.base import BaseOutput
from vellum.workflows.state.context import WorkflowContext, WorkflowDeploymentMetadata
from vellum.workflows.types.core import EntityInputsInterface, MergeBehavior
from vellum.workflows.types.generics import StateType
from vellum.workflows.workflows.event_filters import all_workflow_event_filter

if TYPE_CHECKING:
    from vellum.workflows.workflows.base import BaseWorkflow


class SubworkflowDeploymentNode(BaseNode[StateType], Generic[StateType]):
    """
    Used to execute a Workflow Deployment.

    subworkflow_inputs: EntityInputsInterface - The inputs for the Subworkflow
    deployment: Union[UUID, str] - Either the Workflow Deployment's UUID or its name.
    release_tag: str = LATEST_RELEASE_TAG - The release tag to use for the Workflow Execution
    external_id: Optional[str] = OMIT - Optionally include a unique identifier for tracking purposes.
        Must be unique within a given Workflow Deployment.
    expand_meta: Optional[WorkflowExpandMetaRequest] = OMIT - Expandable execution fields to include in the response
    metadata: Optional[Dict[str, Optional[Any]]] = OMIT - The metadata to use for the Workflow Execution
    request_options: Optional[RequestOptions] = None - The request options to use for the Workflow Execution
    """

    # Either the Workflow Deployment's UUID or its name.
    deployment: ClassVar[Union[UUID, str]]
    subworkflow_inputs: ClassVar[Union[EntityInputsInterface, BaseInputs]] = {}

    release_tag: str = LATEST_RELEASE_TAG
    external_id: Optional[str] = OMIT

    expand_meta: Optional[WorkflowExpandMetaRequest] = OMIT
    metadata: Optional[Dict[str, Optional[Any]]] = OMIT

    request_options: Optional[RequestOptions] = None

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    def _compile_subworkflow_inputs(self) -> List[WorkflowRequestInputRequest]:
        # TODO: We may want to consolidate with prompt deployment input compilation
        # https://app.shortcut.com/vellum/story/4117

        compiled_inputs: List[WorkflowRequestInputRequest] = []

        if isinstance(self.subworkflow_inputs, BaseInputs):
            for input_descriptor, input_value in self.subworkflow_inputs:
                self._add_compiled_input(compiled_inputs, input_descriptor.name, input_value)
        else:
            for input_name, input_value in self.subworkflow_inputs.items():
                self._add_compiled_input(compiled_inputs, input_name, input_value)

        return compiled_inputs

    def _add_compiled_input(
        self, compiled_inputs: List[WorkflowRequestInputRequest], input_name: str, input_value: Any
    ) -> None:
        # Exclude inputs that resolved to be null. This ensure that we don't pass input values
        # to optional subworkflow inputs whose values were unresolved.
        if input_value is None:
            return
        if isinstance(input_value, str):
            compiled_inputs.append(
                WorkflowRequestStringInputRequest(
                    name=input_name,
                    value=input_value,
                )
            )
        elif (
            isinstance(input_value, list)
            and len(input_value) > 0
            and all(isinstance(message, (ChatMessage, ChatMessageRequest)) for message in input_value)
        ):
            chat_history = [
                (
                    message
                    if isinstance(message, ChatMessageRequest)
                    else ChatMessageRequest.model_validate(message.model_dump())
                )
                for message in input_value
                if isinstance(message, (ChatMessage, ChatMessageRequest))
            ]
            compiled_inputs.append(
                WorkflowRequestChatHistoryInputRequest(
                    name=input_name,
                    value=chat_history,
                )
            )
        elif isinstance(input_value, dict):
            compiled_inputs.append(
                WorkflowRequestJsonInputRequest(
                    name=input_name,
                    value=cast(Dict[str, Any], input_value),
                )
            )
        elif isinstance(input_value, (int, float)):
            compiled_inputs.append(
                WorkflowRequestNumberInputRequest(
                    name=input_name,
                    value=input_value,
                )
            )
        else:
            try:
                input_value = default_serializer(input_value)
            except json.JSONDecodeError as e:
                raise NodeException(
                    message=f"Failed to serialize input '{input_name}' of type '{input_value.__class__}': {e}",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                )
            compiled_inputs.append(
                WorkflowRequestJsonInputRequest(
                    name=input_name,
                    value=input_value,
                )
            )

    def _compile_subworkflow_inputs_for_direct_invocation(self, workflow: "BaseWorkflow") -> Any:
        """Compile inputs for direct workflow invocation (similar to InlineSubworkflowNode)."""
        inputs_class = workflow.get_inputs_class()

        if isinstance(self.subworkflow_inputs, BaseInputs):
            inputs_dict = {}
            for input_descriptor, input_value in self.subworkflow_inputs:
                if input_value is not None:
                    inputs_dict[input_descriptor.name] = input_value
            return inputs_class(**inputs_dict)
        else:
            # Filter out None values for direct invocation
            filtered_inputs = {k: v for k, v in self.subworkflow_inputs.items() if v is not None}
            return inputs_class(**filtered_inputs)

    def _run_resolved_workflow(
        self,
        workflow_class: Type["BaseWorkflow"],
        deployment_metadata: Optional[WorkflowDeploymentMetadata],
    ) -> Iterator[BaseOutput]:
        """Execute resolved workflow directly (similar to InlineSubworkflowNode)."""
        # Construct the parent context hierarchy for the subworkflow
        parent_context = get_parent_context()

        # If we have deployment metadata, wrap the parent context with WorkflowDeploymentParentContext
        if deployment_metadata:
            parent_context = WorkflowDeploymentParentContext(
                span_id=uuid4(),
                deployment_id=deployment_metadata.deployment_id,
                deployment_name=deployment_metadata.deployment_name,
                deployment_history_item_id=deployment_metadata.deployment_history_item_id,
                release_tag_id=deployment_metadata.release_tag_id,
                release_tag_name=deployment_metadata.release_tag_name,
                workflow_version_id=deployment_metadata.workflow_version_id,
                external_id=self.external_id if self.external_id is not OMIT else None,
                metadata=self.metadata if self.metadata is not OMIT else None,
                parent=parent_context,
            )

        with execution_context(parent_context=parent_context):
            # Instantiate the workflow inside the execution context so it captures the correct parent context
            resolved_workflow = workflow_class(
                context=WorkflowContext.create_from(self._context), parent_state=self.state
            )
            subworkflow_stream = resolved_workflow.stream(
                inputs=self._compile_subworkflow_inputs_for_direct_invocation(resolved_workflow),
                event_filter=all_workflow_event_filter,
                node_output_mocks=self._context._get_all_node_output_mocks(),
            )

            try:
                first_event = next(subworkflow_stream)
                self._context._emit_subworkflow_event(first_event)
            except WorkflowInitializationException as e:
                hashed_module = e.definition.__module__
                raise NodeException(
                    message=e.message,
                    code=e.code,
                    raw_data={"hashed_module": hashed_module},
                ) from e

            outputs = None
            exception = None
            fulfilled_output_names: Set[str] = set()

            for event in subworkflow_stream:
                self._context._emit_subworkflow_event(event)
                if exception:
                    continue

                if not is_workflow_event(event):
                    continue
                if event.workflow_definition != resolved_workflow.__class__:
                    continue

                if event.name == "workflow.execution.streaming":
                    if event.output.is_fulfilled:
                        fulfilled_output_names.add(event.output.name)
                    yield event.output
                elif event.name == "workflow.execution.fulfilled":
                    outputs = event.outputs
                elif event.name == "workflow.execution.rejected":
                    exception = NodeException.of(event.error)
                elif event.name == "workflow.execution.paused":
                    exception = NodeException(
                        code=WorkflowErrorCode.INVALID_OUTPUTS,
                        message="Subworkflow unexpectedly paused",
                    )

            if exception:
                raise exception

            if outputs is None:
                raise NodeException(
                    message="Expected to receive outputs from Workflow Deployment",
                    code=WorkflowErrorCode.INVALID_OUTPUTS,
                )

            for output_descriptor, output_value in outputs:
                if output_descriptor.name not in fulfilled_output_names:
                    yield BaseOutput(
                        name=output_descriptor.name,
                        value=output_value,
                    )

    def run(self) -> Iterator[BaseOutput]:
        execution_context = get_execution_context()
        request_options = self.request_options or RequestOptions()
        request_options["additional_body_parameters"] = {
            "execution_context": execution_context.model_dump(mode="json"),
            **request_options.get("additional_body_parameters", {}),
        }

        if self.deployment is undefined:
            raise NodeException(
                code=WorkflowErrorCode.NODE_EXECUTION,
                message="Expected subworkflow deployment attribute to be either a UUID or STR, got `undefined` instead",
            )

        try:
            deployment_id = str(self.deployment) if isinstance(self.deployment, UUID) else None
            deployment_name = self.deployment if isinstance(self.deployment, str) else None
        except AttributeError:
            raise NodeException(
                code=WorkflowErrorCode.NODE_EXECUTION,
                message="Expected subworkflow deployment attribute to be either a UUID or STR, got None instead",
            )

        if not deployment_name:
            raise NodeException(
                code=WorkflowErrorCode.INVALID_INPUTS,
                message="Expected deployment name to be provided for subworkflow execution.",
            )

        resolved_result = self._context.resolve_workflow_deployment(
            deployment_name=deployment_name, release_tag=self.release_tag, state=self.state
        )
        if resolved_result:
            workflow_class, deployment_metadata = resolved_result
            yield from self._run_resolved_workflow(workflow_class, deployment_metadata)
            return

        try:
            subworkflow_stream = self._context.vellum_client.execute_workflow_stream(
                inputs=self._compile_subworkflow_inputs(),
                workflow_deployment_id=deployment_id,
                workflow_deployment_name=deployment_name,
                release_tag=self.release_tag,
                external_id=self.external_id,
                event_types=["WORKFLOW"],
                metadata=self.metadata,
                request_options=request_options,
            )
        except ApiError as e:
            self._handle_api_error(e)

        # We don't use the INITIATED event anyway, so we can just skip it
        # and use the exception handling to catch other api level errors
        try:
            next(subworkflow_stream)
        except ApiError as e:
            self._handle_api_error(e)

        outputs: Optional[List[WorkflowOutput]] = None
        fulfilled_output_names: Set[str] = set()
        for event in subworkflow_stream:
            if event.type != "WORKFLOW":
                continue
            if event.data.state == "INITIATED":
                continue
            elif event.data.state == "STREAMING":
                if event.data.output:
                    if event.data.output.state == "STREAMING":
                        yield BaseOutput(
                            name=event.data.output.name,
                            delta=event.data.output.delta,
                        )
                    elif event.data.output.state == "FULFILLED":
                        yield BaseOutput(
                            name=event.data.output.name,
                            value=event.data.output.value,
                        )
                        fulfilled_output_names.add(event.data.output.name)
            elif event.data.state == "FULFILLED":
                outputs = event.data.outputs
            elif event.data.state == "REJECTED":
                error = event.data.error
                if not error:
                    raise NodeException(
                        message="Expected to receive an error from REJECTED event",
                        code=WorkflowErrorCode.INTERNAL_ERROR,
                    )
                workflow_error = workflow_event_error_to_workflow_error(error)
                raise NodeException.of(workflow_error)

        if outputs is None:
            raise NodeException(
                message="Expected to receive outputs from Workflow Deployment",
                code=WorkflowErrorCode.INTERNAL_ERROR,
            )

        # For any outputs somehow in our final fulfilled outputs array,
        # but not fulfilled by the stream.
        for output in outputs:
            if output.name not in fulfilled_output_names:
                yield BaseOutput(
                    name=output.name,
                    value=output.value,
                )

    def _handle_api_error(self, e: ApiError):
        if e.status_code and e.status_code >= 400 and e.status_code < 500 and isinstance(e.body, dict):
            raise NodeException(
                message=e.body.get("detail", "Failed to execute Subworkflow Deployment"),
                code=WorkflowErrorCode.INVALID_INPUTS,
            ) from e

        raise NodeException(
            message="Failed to execute Subworkflow Deployment",
            code=WorkflowErrorCode.INTERNAL_ERROR,
        ) from e



from typing import Generic, Set

from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.ports.port import Port
from vellum.workflows.ports.utils import validate_ports
from vellum.workflows.state.base import BaseState
from vellum.workflows.types import MergeBehavior
from vellum.workflows.types.generics import StateType


class ConditionalNode(BaseNode[StateType], Generic[StateType]):
    """
    Used to conditionally determine which port to invoke next. This node exists to be backwards compatible with
    Vellum's Conditional Node, and for most cases, you should extend `BaseNode.Ports` directly.
    """

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    class Ports(BaseNode.Ports):
        def __call__(self, outputs: BaseOutputs, state: BaseState) -> Set[Port]:
            all_ports = [port for port in self.__class__]
            enforce_single_invoked_port = validate_ports(all_ports)

            if not enforce_single_invoked_port:
                raise ValueError("Conditional nodes must have exactly one if port")

            return super().__call__(outputs, state)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .prompt_execution_meta import PromptExecutionMeta
from .prompt_output import PromptOutput


class FulfilledExecutePromptResponse(UniversalBaseModel):
    """
    The successful response from the model containing all of the resolved values generated by the prompt.
    """

    meta: typing.Optional[PromptExecutionMeta] = None
    raw: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = pydantic.Field(default=None)
    """
    The subset of the raw response from the model that the request opted into with `expand_raw`.
    """

    execution_id: str = pydantic.Field()
    """
    The ID of the execution.
    """

    state: typing.Literal["FULFILLED"] = "FULFILLED"
    outputs: typing.List[PromptOutput]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from typing import Generic

from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.types import MergeBehavior
from vellum.workflows.types.generics import StateType


class MergeNode(BaseNode[StateType], Generic[StateType]):
    """
    Used to merge the control flow of multiple nodes into a single node. This node exists to be backwards compatible
    with Vellum's Merge Node, and for most cases, you should extend from `BaseNode.Trigger` directly.
    """

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY



# This file was auto-generated by Fern from our API Definition.

import typing

from .workflow_event import WorkflowEvent

CreateWorkflowEventRequest = typing.Union[typing.List[WorkflowEvent], WorkflowEvent]



from typing import NamedTuple

from vellum.workflows.state.base import BaseState


class LoadStateResult(NamedTuple):
    state: BaseState
    previous_trace_id: str
    previous_span_id: str
    root_trace_id: str
    root_span_id: str



from typing import List

from vellum import ChatMessage
from vellum.workflows.state.base import BaseState


class ToolCallingState(BaseState):
    chat_history: List[ChatMessage] = []
    prompt_iterations: int = 0
    current_prompt_output_index: int = 0
    current_function_calls_processed: int = 0



"""Tests for BaseTrigger.Display serialization."""

from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.triggers.schedule import ScheduleTrigger
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_serialize_basetrigger_with_display_class():
    """Tests that BaseTrigger.Display class attributes serialize correctly."""

    class MyTrigger(ScheduleTrigger):
        class Config:
            cron = "0 9 * * *"
            timezone = "UTC"

        class Display:
            icon = "vellum:icon:gear"
            color = "purple"

    class SimpleNode(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = MyTrigger >> SimpleNode

    # WHEN we serialize the workflow
    result: dict = get_workflow_display(workflow_class=TestWorkflow).serialize()

    assert "triggers" in result
    triggers = result["triggers"]
    assert isinstance(triggers, list)
    assert len(triggers) == 1

    trigger = triggers[0]
    assert isinstance(trigger, dict)
    assert "display_data" in trigger
    display_data = trigger["display_data"]
    assert isinstance(display_data, dict)
    assert display_data["icon"] == "vellum:icon:gear"
    assert display_data["color"] == "purple"


def test_serialize_trigger_with_only_icon():
    """Tests that triggers with only icon serialize correctly."""

    class IconOnlyTrigger(ScheduleTrigger):
        class Config:
            cron = "0 9 * * *"
            timezone = "UTC"

        class Display:
            icon = "vellum:icon:star"

    class SimpleNode(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = IconOnlyTrigger >> SimpleNode

    # WHEN we serialize the workflow
    result: dict = get_workflow_display(workflow_class=TestWorkflow).serialize()

    triggers = result["triggers"]
    assert len(triggers) == 1
    trigger = triggers[0]
    display_data = trigger["display_data"]
    assert display_data["icon"] == "vellum:icon:star"
    assert "color" not in display_data


def test_serialize_trigger_with_only_color():
    """Tests that triggers with only color serialize correctly."""

    class ColorOnlyTrigger(ScheduleTrigger):
        class Config:
            cron = "0 9 * * *"
            timezone = "UTC"

        class Display:
            color = "#FF5733"

    class SimpleNode(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = ColorOnlyTrigger >> SimpleNode

    # WHEN we serialize the workflow
    result: dict = get_workflow_display(workflow_class=TestWorkflow).serialize()

    triggers = result["triggers"]
    assert len(triggers) == 1
    trigger = triggers[0]
    display_data = trigger["display_data"]
    assert display_data["color"] == "#FF5733"
    assert "icon" not in display_data


def test_serialize_trigger_without_display_class():
    """Tests that triggers without Display class don't have display_data."""

    class PlainTrigger(ScheduleTrigger):
        class Config:
            cron = "0 9 * * *"
            timezone = "UTC"

    class SimpleNode(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = PlainTrigger >> SimpleNode

    # WHEN we serialize the workflow
    result: dict = get_workflow_display(workflow_class=TestWorkflow).serialize()

    triggers = result["triggers"]
    assert len(triggers) == 1
    trigger = triggers[0]
    display_data = trigger.get("display_data")
    if display_data is not None:
        assert display_data == {}


def test_serialize_trigger_with_none_values():
    """Tests that triggers with None icon/color don't serialize those fields."""

    class NoneValuesTrigger(ScheduleTrigger):
        class Config:
            cron = "0 9 * * *"
            timezone = "UTC"

        class Display:
            icon = None
            color = None

    class SimpleNode(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NoneValuesTrigger >> SimpleNode

    # WHEN we serialize the workflow
    result: dict = get_workflow_display(workflow_class=TestWorkflow).serialize()

    triggers = result["triggers"]
    assert len(triggers) == 1
    trigger = triggers[0]
    display_data = trigger.get("display_data")
    if display_data is not None:
        assert display_data == {}



# This file was auto-generated by Fern from our API Definition.

import typing

VellumVariableType = typing.Union[
    typing.Literal[
        "STRING",
        "NUMBER",
        "JSON",
        "CHAT_HISTORY",
        "SEARCH_RESULTS",
        "ERROR",
        "ARRAY",
        "FUNCTION_CALL",
        "AUDIO",
        "VIDEO",
        "IMAGE",
        "DOCUMENT",
        "NULL",
        "THINKING",
    ],
    typing.Any,
]



import importlib
from importlib.machinery import ModuleSpec
from io import StringIO
import re
import sys
from typing import Optional

from vellum.workflows.loaders.base import BaseWorkflowFinder


class VirtualFileLoader(importlib.abc.Loader):
    def __init__(self, files: dict[str, str], namespace: str, source_module: Optional[str] = None):
        self.files = files
        self.namespace = namespace
        self.source_module = source_module

    def create_module(self, spec: ModuleSpec):
        """
        We started with cpython/Lib/importlib/_bootstrap.py::FrozenImporter::create_module here

        https://github.com/python/cpython/blob/053c285f6b41f92fbdd1d4ff0c959cceefacd7cd/Lib/importlib/_bootstrap.py#L1160C1-L1169C22

        and reduced our needs to just updating the __file__ attribute directly.
        """
        module = type(sys)(spec.name)
        module.__file__ = spec.origin
        return module

    def exec_module(self, module):
        module_info = self._resolve_module(module.__spec__.name)

        if module_info:
            file_path, code = module_info
            namespaced_path = (
                file_path if file_path.startswith(f"{self.namespace}/") else f"{self.namespace}/{file_path}"
            )
            module.__file__ = namespaced_path
            compiled = compile(code, namespaced_path, "exec")
            exec(compiled, module.__dict__)

    def get_source(self, fullname):
        """
        `inspect` module uses this method to get the source code of a module.
        """

        module_info = self._resolve_module(fullname)
        if module_info:
            return module_info[1]

        return None

    def _resolve_module(self, fullname: str) -> Optional[tuple[str, str]]:
        file_path = self._get_file_path(fullname)
        code = self._get_code(file_path)

        if code is not None:
            return file_path, code

        if not file_path.endswith("__init__.py"):
            file_path = re.sub(r"\.py$", "/__init__.py", file_path)
            code = self._get_code(file_path)

            if code is not None:
                return file_path, code

            relative_name = self._to_relative(fullname)
            if relative_name is not None and self._is_package_directory(relative_name):
                return self._generate_init_content(relative_name)

        return None

    def _to_relative(self, fullname: str) -> Optional[str]:
        """Convert a fully qualified module name to a relative name without the namespace prefix."""
        if self.source_module:
            if fullname.startswith(self.source_module + "."):
                return fullname[len(self.source_module) + 1 :]
            elif fullname == self.source_module:
                return ""

        if fullname.startswith(self.namespace + "."):
            return fullname[len(self.namespace) + 1 :]
        elif fullname == self.namespace:
            return ""
        return None

    def _get_file_path(self, fullname):
        return f"{fullname.replace('.', '/')}.py"

    def _get_code(self, file_path):
        if self.source_module:
            source_module_prefix = re.escape(self.source_module) + r"/"
            file_key_name = re.sub(r"^" + source_module_prefix, "", file_path)
            if file_key_name != file_path:
                return self.files.get(file_key_name)

        file_key_name = re.sub(r"^" + re.escape(self.namespace) + r"/", "", file_path)
        return self.files.get(file_key_name)

    def _is_package_directory(self, fullname: str) -> bool:
        """Check if directory contains .py files that should be treated as a package."""
        if fullname == "":
            # This is the root namespace, so it's a package directory
            return True

        directory_prefix = fullname.replace(".", "/") + "/"
        for file_path in self.files.keys():
            if file_path.startswith(directory_prefix):
                if file_path.endswith(".py") and not file_path.endswith("__init__.py"):
                    return True
                remaining_path = file_path[len(directory_prefix) :]
                if "/" in remaining_path:
                    return True

        return False

    def _generate_init_content(self, fullname: str) -> tuple[str, str]:
        """Auto-generate empty __init__.py content to mark directory as a package."""
        if fullname == "":
            if any(file_path.startswith("display/") for file_path in self.files.keys()):
                return (
                    "__init__.py",
                    """\
from .display import *
""",
                )
            else:
                return "__init__.py", ""

        directory_prefix = fullname.replace(".", "/") + "/"
        file_path = directory_prefix + "__init__.py"

        if not fullname.startswith("display"):
            return "__init__.py", ""

        # We have to go through the effort of exhaustively importing all display modules, because that's how
        # dynamically annotate our Workflow graph.
        all_nested_files = [
            file_path[len(directory_prefix) :]
            for file_path in self.files.keys()
            if file_path.startswith(directory_prefix)
        ]

        top_level_items = {file_path.split("/")[0] for file_path in all_nested_files}

        top_level_modules = [
            re.sub(r"\.py$", "", file_path)
            for file_path in top_level_items
            if file_path.endswith(".py") or "." not in file_path
        ]

        code = "\n".join([f"from .{file_path} import *" for file_path in top_level_modules])

        return file_path, code

    def virtual_open(self, file_path: str) -> Optional[StringIO]:
        """
        Open a virtual file if it exists in this loader's namespace.

        Args:
            file_path: The absolute file path to open

        Returns:
            A StringIO object containing the file contents, or None if the file is not found
        """
        if file_path.startswith(self.namespace + "/"):
            relative_path = file_path[len(self.namespace) + 1 :]
            content = self._get_code(relative_path)
            if content is not None:
                return StringIO(content)
        return None


class VirtualFileFinder(BaseWorkflowFinder):
    def __init__(self, files: dict[str, str], namespace: str, source_module: Optional[str] = None):
        self.loader = VirtualFileLoader(files, namespace, source_module)
        self.source_module = source_module
        self.namespace = namespace

    def format_error_message(self, error_message: str) -> str:
        """Format error message by replacing namespace with source_module."""
        if self.source_module and self.namespace in error_message:
            return error_message.replace(self.namespace, self.source_module)
        return error_message

    def virtual_open(self, file_path: str) -> Optional[StringIO]:
        """
        Open a virtual file by delegating to the loader.

        Args:
            file_path: The absolute file path to open

        Returns:
            A StringIO object containing the file contents, or None if the file is not found
        """
        return self.loader.virtual_open(file_path)

    def find_spec(self, fullname: str, path, target=None):
        module_info = self.loader._resolve_module(fullname)
        if isinstance(module_info, tuple) and len(module_info) == 2:
            file_path, _ = module_info
            is_package = file_path.endswith("__init__.py")
            return importlib.machinery.ModuleSpec(
                fullname,
                self.loader,
                origin=file_path,
                is_package=is_package,
            )

        return None



from abc import abstractmethod
from itertools import chain
from typing import ClassVar, Generator, Generic, Iterator, List, Optional, Union, cast

from vellum import AdHocExecutePromptEvent, ExecutePromptEvent, PromptOutput
from vellum.client.core import RequestOptions
from vellum.client.core.api_error import ApiError
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode, vellum_error_to_workflow_error
from vellum.workflows.events.node import NodeExecutionStreamingEvent
from vellum.workflows.exceptions import NodeException
from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs.base import BaseOutput, BaseOutputs
from vellum.workflows.references.output import OutputReference
from vellum.workflows.types.core import EntityInputsInterface, MergeBehavior
from vellum.workflows.types.generics import StateType


def _contains_reference_to_output(reference: BaseDescriptor, target_reference: OutputReference) -> bool:
    if reference == target_reference:
        return True
    if isinstance(reference, CoalesceExpression):
        return _contains_reference_to_output(reference.lhs, target_reference) or _contains_reference_to_output(
            reference.rhs, target_reference
        )
    return False


class BasePromptNode(BaseNode[StateType], Generic[StateType]):
    # Inputs that are passed to the Prompt
    prompt_inputs: ClassVar[Optional[EntityInputsInterface]] = None

    request_options: Optional[RequestOptions] = None

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    class Outputs(BaseOutputs):
        results: List[PromptOutput]

    @abstractmethod
    def _get_prompt_event_stream(self) -> Union[Iterator[AdHocExecutePromptEvent], Iterator[ExecutePromptEvent]]:
        pass

    def _validate(self) -> None:
        pass

    def run(self) -> Iterator[BaseOutput]:
        outputs = yield from self._process_prompt_event_stream()
        if outputs is None:
            raise NodeException(
                message="Expected to receive outputs from Prompt",
                code=WorkflowErrorCode.INTERNAL_ERROR,
            )

    def _process_prompt_event_stream(self) -> Generator[BaseOutput, None, Optional[List[PromptOutput]]]:
        self._validate()
        try:
            prompt_event_stream = self._get_prompt_event_stream()
        except ApiError as e:
            self._handle_api_error(e)

        try:
            first_event = next(prompt_event_stream)
        except ApiError as e:
            self._handle_api_error(e)
        else:
            if first_event.state == "REJECTED":
                workflow_error = vellum_error_to_workflow_error(first_event.error)
                raise NodeException.of(workflow_error)
            if first_event.state != "INITIATED":
                prompt_event_stream = cast(
                    Union[Iterator[AdHocExecutePromptEvent], Iterator[ExecutePromptEvent]],
                    chain([first_event], prompt_event_stream),
                )

        outputs: Optional[List[PromptOutput]] = None
        exception: Optional[NodeException] = None
        for event in prompt_event_stream:
            if exception:
                continue

            if event.state == "INITIATED":
                continue
            elif event.state == "STREAMING":
                yield BaseOutput(name="results", delta=event.output.value)
            elif event.state == "FULFILLED":
                outputs = event.outputs
                yield BaseOutput(name="results", value=event.outputs)
            elif event.state == "REJECTED":
                workflow_error = vellum_error_to_workflow_error(event.error)
                exception = NodeException.of(workflow_error)

        if exception:
            raise exception

        return outputs

    def _handle_api_error(self, e: ApiError):
        if e.status_code and e.status_code == 403 and isinstance(e.body, dict):
            raise NodeException(
                message=e.body.get("detail", "Provider credentials is missing or unavailable"),
                code=WorkflowErrorCode.PROVIDER_CREDENTIALS_UNAVAILABLE,
                raw_data=e.body,
            )

        elif e.status_code and e.status_code >= 400 and e.status_code < 500 and isinstance(e.body, dict):
            raise NodeException(
                message=e.body.get("detail", "Failed to execute Prompt"),
                code=WorkflowErrorCode.INVALID_INPUTS,
                raw_data=e.body,
            ) from e

        raise NodeException(
            message="Failed to execute Prompt",
            code=WorkflowErrorCode.INTERNAL_ERROR,
            raw_data=e.body,
        ) from e

    def __directly_emit_workflow_output__(
        self,
        event: NodeExecutionStreamingEvent,
        workflow_output_descriptor: OutputReference,
    ) -> bool:
        if event.output.name != "results":
            return False

        if not isinstance(event.output.delta, str) and not event.output.is_initiated:
            return False

        target_nodes = [e.to_node for port in self.Ports for e in port.edges if e.to_node.__simulates_workflow_output__]
        target_node_output = next(
            (
                o
                for target_node in target_nodes
                for o in target_node.Outputs
                if o == workflow_output_descriptor.instance
            ),
            None,
        )
        if not target_node_output:
            return False

        if not isinstance(target_node_output.instance, BaseDescriptor):
            return False

        return _contains_reference_to_output(target_node_output.instance, event.node_definition.Outputs.text)
